# Copyright 2015 Hewlett Packard Enterprise Development LP
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy
import sys

import ddt
import mock
if 'hpe3parclient' not in sys.modules:
    sys.modules['hpe3parclient'] = mock.Mock()

from manila import exception
from manila.share.drivers.hpe import hpe_3par_driver as hpe3pardriver
from manila.share.drivers.hpe import hpe_3par_mediator as hpe3parmediator
from manila import test
from manila.tests.share.drivers.hpe import test_hpe_3par_constants as constants


@ddt.ddt
class HPE3ParDriverFPGTestCase(test.TestCase):

    @ddt.data((-1, 4),
              (0, 5),
              (0, -1))
    @ddt.unpack
    def test_FPG_init_args_failure(self, min_ip, max_ip):
        self.assertRaises(exception.HPE3ParInvalid,
                          hpe3pardriver.FPG, min_ip, max_ip)

    @ddt.data(('invalid_ip_fpg, 10.256.0.1', 0, 4),
              (None, 0, 4),
              (' ', 0, 4),
              ('', 0, 4),
              ('max_ip_fpg, 10.0.0.1, 10.0.0.2, 10.0.0.3, 10.0.0.4, 10.0.0.5',
              0, 4),
              ('min_1_ip_fpg', 1, 4))
    @ddt.unpack
    def test_FPG_type_failures(self, value, min_ip, max_ip):
        fpg_type_obj = hpe3pardriver.FPG(min_ip=min_ip, max_ip=max_ip)
        self.assertRaises(exception.HPE3ParInvalid, fpg_type_obj, value)

    @ddt.data(('samplefpg, 10.0.0.1', {'samplefpg': ['10.0.0.1']}),
              ('samplefpg', {'samplefpg': []}),
              ('samplefpg, 10.0.0.1, 10.0.0.2',
               {'samplefpg': ['10.0.0.1', '10.0.0.2']}))
    @ddt.unpack
    def test_FPG_type_success(self, value, expected_fpg):
        fpg_type_obj = hpe3pardriver.FPG()
        fpg = fpg_type_obj(value)
        self.assertEqual(expected_fpg, fpg)


@ddt.ddt
class HPE3ParDriverTestCase(test.TestCase):

    def setUp(self):
        super(HPE3ParDriverTestCase, self).setUp()

        # Create a mock configuration with attributes and a safe_get()
        self.conf = mock.Mock()
        self.conf.driver_handles_share_servers = True
        self.conf.hpe3par_debug = constants.EXPECTED_HPE_DEBUG
        self.conf.hpe3par_username = constants.USERNAME
        self.conf.hpe3par_password = constants.PASSWORD
        self.conf.hpe3par_api_url = constants.API_URL
        self.conf.hpe3par_san_login = constants.SAN_LOGIN
        self.conf.hpe3par_san_password = constants.SAN_PASSWORD
        self.conf.hpe3par_san_ip = constants.EXPECTED_IP_1234
        self.conf.hpe3par_fpg = constants.EXPECTED_FPG_CONF
        self.conf.hpe3par_san_ssh_port = constants.PORT
        self.conf.ssh_conn_timeout = constants.TIMEOUT
        self.conf.hpe3par_fstore_per_share = False
        self.conf.hpe3par_require_cifs_ip = False
        self.conf.hpe3par_cifs_admin_access_username = constants.USERNAME,
        self.conf.hpe3par_cifs_admin_access_password = constants.PASSWORD,
        self.conf.hpe3par_cifs_admin_access_domain = (
            constants.EXPECTED_CIFS_DOMAIN),
        self.conf.hpe3par_share_mount_path = constants.EXPECTED_MOUNT_PATH,
        self.conf.my_ip = constants.EXPECTED_IP_1234
        self.conf.network_config_group = 'test_network_config_group'
        self.conf.admin_network_config_group = (
            'test_admin_network_config_group')
        self.conf.filter_function = None
        self.conf.goodness_function = None

        def safe_get(attr):
            try:
                return self.conf.__getattribute__(attr)
            except AttributeError:
                return None
        self.conf.safe_get = safe_get

        self.real_hpe_3par_mediator = hpe3parmediator.HPE3ParMediator
        self.mock_object(hpe3parmediator, 'HPE3ParMediator')
        self.mock_mediator_constructor = hpe3parmediator.HPE3ParMediator
        self.mock_mediator = self.mock_mediator_constructor()
        # restore needed static methods
        self.mock_mediator.ensure_supported_protocol = (
            self.real_hpe_3par_mediator.ensure_supported_protocol)
        self.mock_mediator.build_export_locations = (
            self.real_hpe_3par_mediator.build_export_locations)

        self.driver = hpe3pardriver.HPE3ParShareDriver(
            configuration=self.conf)

    def test_driver_setup_success(self,
                                  get_vfs_ret_val=constants.EXPECTED_GET_VFS):
        """Driver do_setup without any errors."""

        self.mock_mediator.get_vfs.return_value = get_vfs_ret_val

        self.driver.do_setup(None)
        conf = self.conf
        self.mock_mediator_constructor.assert_has_calls([
            mock.call(hpe3par_san_ssh_port=conf.hpe3par_san_ssh_port,
                      hpe3par_san_password=conf.hpe3par_san_password,
                      hpe3par_username=conf.hpe3par_username,
                      hpe3par_san_login=conf.hpe3par_san_login,
                      hpe3par_debug=conf.hpe3par_debug,
                      hpe3par_api_url=conf.hpe3par_api_url,
                      hpe3par_password=conf.hpe3par_password,
                      hpe3par_san_ip=conf.hpe3par_san_ip,
                      hpe3par_fstore_per_share=conf.hpe3par_fstore_per_share,
                      hpe3par_require_cifs_ip=conf.hpe3par_require_cifs_ip,
                      hpe3par_cifs_admin_access_username=(
                          conf.hpe3par_cifs_admin_access_username),
                      hpe3par_cifs_admin_access_password=(
                          conf.hpe3par_cifs_admin_access_password),
                      hpe3par_cifs_admin_access_domain=(
                          conf.hpe3par_cifs_admin_access_domain),
                      hpe3par_share_mount_path=conf.hpe3par_share_mount_path,
                      my_ip=self.conf.my_ip,
                      ssh_conn_timeout=conf.ssh_conn_timeout)])

        self.mock_mediator.assert_has_calls([
            mock.call.do_setup(),
            mock.call.get_vfs(constants.EXPECTED_FPG)])

    def test_driver_setup_dhss_success(self):
        """Driver do_setup without any errors with dhss=True."""

        self.test_driver_setup_success()
        self.assertEqual(constants.EXPECTED_FPG_MAP, self.driver.fpgs)

    def test_driver_setup_no_dhss_success(self):
        """Driver do_setup without any errors with dhss=False."""

        self.conf.driver_handles_share_servers = False
        self.test_driver_setup_success()
        self.assertEqual(constants.EXPECTED_FPG_MAP, self.driver.fpgs)

    def test_driver_setup_no_dhss_multi_getvfs_success(self):
        """Driver do_setup when dhss=False, getvfs returns multiple IPs."""

        self.conf.driver_handles_share_servers = False
        self.test_driver_setup_success(
            get_vfs_ret_val=constants.EXPECTED_GET_VFS_MULTIPLES)
        self.assertEqual(constants.EXPECTED_FPG_MAP,
                         self.driver.fpgs)

    def test_driver_setup_success_no_dhss_no_conf_ss_ip(self):
        """test driver's do_setup()

        Driver do_setup with dhss=False, share server ip not set in config file
        but discoverable at 3par array
        """

        self.conf.driver_handles_share_servers = False
        # ss ip not provided in conf
        original_fpg = deepcopy(self.conf.hpe3par_fpg)
        self.conf.hpe3par_fpg[0][constants.EXPECTED_FPG] = []

        self.test_driver_setup_success()

        self.assertEqual(constants.EXPECTED_FPG_MAP, self.driver.fpgs)
        constants.EXPECTED_FPG_CONF = original_fpg

    def test_driver_setup_failure_no_dhss_no_conf_ss_ip(self):
        """Configured IP address is required for dhss=False."""

        self.conf.driver_handles_share_servers = False
        # ss ip not provided in conf
        fpg_without_ss_ip = deepcopy(self.conf.hpe3par_fpg)
        self.conf.hpe3par_fpg[0][constants.EXPECTED_FPG] = []
        # ss ip not configured on array
        vfs_without_ss_ip = deepcopy(constants.EXPECTED_GET_VFS)
        vfs_without_ss_ip['vfsip']['address'] = []
        self.mock_mediator.get_vfs.return_value = vfs_without_ss_ip

        self.assertRaises(exception.HPE3ParInvalid,
                          self.driver.do_setup, None)
        constants.EXPECTED_FPG_CONF = fpg_without_ss_ip

    def test_driver_setup_mediator_error(self):
        """Driver do_setup when the mediator setup fails."""

        self.mock_mediator.do_setup.side_effect = (
            exception.ShareBackendException('fail'))

        self.assertRaises(exception.ShareBackendException,
                          self.driver.do_setup, None)

        conf = self.conf
        self.mock_mediator_constructor.assert_has_calls([
            mock.call(hpe3par_san_ssh_port=conf.hpe3par_san_ssh_port,
                      hpe3par_san_password=conf.hpe3par_san_password,
                      hpe3par_username=conf.hpe3par_username,
                      hpe3par_san_login=conf.hpe3par_san_login,
                      hpe3par_debug=conf.hpe3par_debug,
                      hpe3par_api_url=conf.hpe3par_api_url,
                      hpe3par_password=conf.hpe3par_password,
                      hpe3par_san_ip=conf.hpe3par_san_ip,
                      hpe3par_fstore_per_share=conf.hpe3par_fstore_per_share,
                      hpe3par_require_cifs_ip=conf.hpe3par_require_cifs_ip,
                      hpe3par_cifs_admin_access_username=(
                          conf.hpe3par_cifs_admin_access_username),
                      hpe3par_cifs_admin_access_password=(
                          conf.hpe3par_cifs_admin_access_password),
                      hpe3par_cifs_admin_access_domain=(
                          conf.hpe3par_cifs_admin_access_domain),
                      hpe3par_share_mount_path=conf.hpe3par_share_mount_path,
                      my_ip=self.conf.my_ip,
                      ssh_conn_timeout=conf.ssh_conn_timeout)])

        self.mock_mediator.assert_has_calls([mock.call.do_setup()])

    def test_driver_setup_with_vfs_error(self):
        """Driver do_setup when the get_vfs fails."""

        self.mock_mediator.get_vfs.side_effect = (
            exception.ShareBackendException('fail'))

        self.assertRaises(exception.ShareBackendException,
                          self.driver.do_setup, None)

        conf = self.conf
        self.mock_mediator_constructor.assert_has_calls([
            mock.call(hpe3par_san_ssh_port=conf.hpe3par_san_ssh_port,
                      hpe3par_san_password=conf.hpe3par_san_password,
                      hpe3par_username=conf.hpe3par_username,
                      hpe3par_san_login=conf.hpe3par_san_login,
                      hpe3par_debug=conf.hpe3par_debug,
                      hpe3par_api_url=conf.hpe3par_api_url,
                      hpe3par_password=conf.hpe3par_password,
                      hpe3par_san_ip=conf.hpe3par_san_ip,
                      hpe3par_fstore_per_share=conf.hpe3par_fstore_per_share,
                      hpe3par_require_cifs_ip=conf.hpe3par_require_cifs_ip,
                      hpe3par_cifs_admin_access_username=(
                          conf.hpe3par_cifs_admin_access_username),
                      hpe3par_cifs_admin_access_password=(
                          conf.hpe3par_cifs_admin_access_password),
                      hpe3par_cifs_admin_access_domain=(
                          conf.hpe3par_cifs_admin_access_domain),
                      hpe3par_share_mount_path=conf.hpe3par_share_mount_path,
                      my_ip=self.conf.my_ip,
                      ssh_conn_timeout=conf.ssh_conn_timeout)])

        self.mock_mediator.assert_has_calls([
            mock.call.do_setup(),
            mock.call.get_vfs(constants.EXPECTED_FPG)])

    def test_driver_setup_conf_ips_validation_fails(self):
        """Driver do_setup when the _validate_pool_ips fails."""

        self.conf.driver_handles_share_servers = False
        vfs_with_ss_ip = deepcopy(constants.EXPECTED_GET_VFS)
        vfs_with_ss_ip['vfsip']['address'] = ['10.100.100.100']
        self.mock_mediator.get_vfs.return_value = vfs_with_ss_ip
        self.assertRaises(exception.HPE3ParInvalid,
                          self.driver.do_setup, None)

        conf = self.conf
        self.mock_mediator_constructor.assert_has_calls([
            mock.call(hpe3par_san_ssh_port=conf.hpe3par_san_ssh_port,
                      hpe3par_san_password=conf.hpe3par_san_password,
                      hpe3par_username=conf.hpe3par_username,
                      hpe3par_san_login=conf.hpe3par_san_login,
                      hpe3par_debug=conf.hpe3par_debug,
                      hpe3par_api_url=conf.hpe3par_api_url,
                      hpe3par_password=conf.hpe3par_password,
                      hpe3par_san_ip=conf.hpe3par_san_ip,
                      hpe3par_fstore_per_share=conf.hpe3par_fstore_per_share,
                      hpe3par_require_cifs_ip=conf.hpe3par_require_cifs_ip,
                      hpe3par_cifs_admin_access_username=(
                          conf.hpe3par_cifs_admin_access_username),
                      hpe3par_cifs_admin_access_password=(
                          conf.hpe3par_cifs_admin_access_password),
                      hpe3par_cifs_admin_access_domain=(
                          conf.hpe3par_cifs_admin_access_domain),
                      hpe3par_share_mount_path=conf.hpe3par_share_mount_path,
                      my_ip=self.conf.my_ip,
                      ssh_conn_timeout=conf.ssh_conn_timeout)])

        self.mock_mediator.assert_has_calls([
            mock.call.do_setup(),
            mock.call.get_vfs(constants.EXPECTED_FPG)])

    def init_driver(self):
        """Simple driver setup for re-use with tests that need one."""

        self.driver._hpe3par = self.mock_mediator
        self.driver.fpgs = constants.EXPECTED_FPG_MAP
        self.mock_object(hpe3pardriver, 'share_types')
        get_extra_specs = hpe3pardriver.share_types.get_extra_specs_from_share
        get_extra_specs.return_value = constants.EXPECTED_EXTRA_SPECS

    def test_driver_check_for_setup_error_success(self):
        """check_for_setup_error when things go well."""

        # Generally this is always mocked, but here we reference the class.
        hpe3parmediator.HPE3ParMediator = self.real_hpe_3par_mediator

        self.mock_object(hpe3pardriver, 'LOG')
        self.init_driver()
        self.driver.check_for_setup_error()
        expected_calls = [
            mock.call.debug('HPE3ParShareDriver SHA1: %s', mock.ANY),
            mock.call.debug('HPE3ParMediator SHA1: %s', mock.ANY)
        ]
        hpe3pardriver.LOG.assert_has_calls(expected_calls)

    def test_driver_check_for_setup_error_exception(self):
        """check_for_setup_error catch and log any exceptions."""

        # Since HPE3ParMediator is mocked, we'll hit the except/log.
        self.mock_object(hpe3pardriver, 'LOG')
        self.init_driver()
        self.driver.check_for_setup_error()
        expected_calls = [
            mock.call.debug('HPE3ParShareDriver SHA1: %s', mock.ANY),
            mock.call.debug('Source code SHA1 not logged due to: %s', mock.ANY)
        ]
        hpe3pardriver.LOG.assert_has_calls(expected_calls)

    @ddt.data(([constants.SHARE_SERVER], constants.SHARE_SERVER),
              ([], None),)
    @ddt.unpack
    def test_choose_share_server_compatible_with_share(self, share_servers,
                                                       expected_share_sever):
        context = None
        share_server = self.driver.choose_share_server_compatible_with_share(
            context,
            share_servers,
            constants.NFS_SHARE_INFO,
            None,
            None)

        self.assertEqual(expected_share_sever, share_server)

    def test_choose_share_server_compatible_with_share_with_cg(self):
        context = None
        cg_ref = {'id': 'dummy'}
        self.assertRaises(
            exception.InvalidRequest,
            self.driver.choose_share_server_compatible_with_share,
            context,
            [constants.SHARE_SERVER],
            constants.NFS_SHARE_INFO,
            None,
            cg_ref)

    def do_create_share(self, protocol, share_type_id, expected_project_id,
                        expected_share_id, expected_size):
        """Re-usable code for create share."""
        context = None

        share = {
            'display_name': constants.EXPECTED_SHARE_NAME,
            'host': constants.EXPECTED_HOST,
            'project_id': expected_project_id,
            'id': expected_share_id,
            'share_proto': protocol,
            'share_type_id': share_type_id,
            'size': expected_size,
        }
        location = self.driver.create_share(context, share,
                                            constants.SHARE_SERVER)
        return location

    def do_create_share_from_snapshot(self,
                                      protocol,
                                      share_type_id,
                                      snapshot_instance,
                                      expected_share_id,
                                      expected_size):
        """Re-usable code for create share from snapshot."""
        context = None
        share = {
            'project_id': constants.EXPECTED_PROJECT_ID,
            'display_name': constants.EXPECTED_SHARE_NAME,
            'host': constants.EXPECTED_HOST,
            'id': expected_share_id,
            'share_proto': protocol,
            'share_type_id': share_type_id,
            'size': expected_size,
        }
        location = self.driver.create_share_from_snapshot(
            context,
            share,
            snapshot_instance,
            constants.SHARE_SERVER)
        return location

    @ddt.data((constants.UNEXPECTED_HOST, exception.InvalidHost),
              (constants.HOST_WITHOUT_POOL_1, exception.InvalidHost),
              (constants.HOST_WITHOUT_POOL_2, exception.InvalidHost))
    @ddt.unpack
    def test_driver_create_share_fails_get_pool_location(self, host,
                                                         expected_exception):
        """get_pool_location fails to extract pool name from host"""
        self.init_driver()
        context = None
        share_server = None
        share = {
            'display_name': constants.EXPECTED_SHARE_NAME,
            'host': host,
            'project_id': constants.EXPECTED_PROJECT_ID,
            'id': constants.EXPECTED_SHARE_ID,
            'share_proto': constants.CIFS,
            'share_type_id': constants.SHARE_TYPE_ID,
            'size': constants.EXPECTED_SIZE_2,
        }
        self.assertRaises(expected_exception,
                          self.driver.create_share,
                          context, share, share_server)

    def test_driver_create_cifs_share(self):
        self.init_driver()

        expected_location = '\\\\%s\%s' % (constants.EXPECTED_IP_10203040,
                                           constants.EXPECTED_SHARE_NAME)

        self.mock_mediator.create_share.return_value = (
            constants.EXPECTED_SHARE_NAME)

        hpe3parmediator.HPE3ParMediator = self.real_hpe_3par_mediator

        location = self.do_create_share(constants.CIFS,
                                        constants.SHARE_TYPE_ID,
                                        constants.EXPECTED_PROJECT_ID,
                                        constants.EXPECTED_SHARE_ID,
                                        constants.EXPECTED_SIZE_2)

        self.assertIn(expected_location, location)
        expected_calls = [mock.call.create_share(
            constants.EXPECTED_PROJECT_ID,
            constants.EXPECTED_SHARE_ID,
            constants.CIFS,
            constants.EXPECTED_EXTRA_SPECS,
            constants.EXPECTED_FPG,
            constants.EXPECTED_VFS,
            comment=mock.ANY,
            size=constants.EXPECTED_SIZE_2)]
        self.mock_mediator.assert_has_calls(expected_calls)

    def test_driver_create_nfs_share(self):
        self.init_driver()

        expected_location = ':'.join((constants.EXPECTED_IP_10203040,
                                      constants.EXPECTED_SHARE_PATH))

        self.mock_mediator.create_share.return_value = (
            constants.EXPECTED_SHARE_PATH)
        hpe3parmediator.HPE3ParMediator = self.real_hpe_3par_mediator

        location = self.do_create_share(constants.NFS,
                                        constants.SHARE_TYPE_ID,
                                        constants.EXPECTED_PROJECT_ID,
                                        constants.EXPECTED_SHARE_ID,
                                        constants.EXPECTED_SIZE_1)

        self.assertIn(expected_location, location)
        expected_calls = [
            mock.call.create_share(constants.EXPECTED_PROJECT_ID,
                                   constants.EXPECTED_SHARE_ID,
                                   constants.NFS,
                                   constants.EXPECTED_EXTRA_SPECS,
                                   constants.EXPECTED_FPG,
                                   constants.EXPECTED_VFS,
                                   comment=mock.ANY,
                                   size=constants.EXPECTED_SIZE_1)]

        self.mock_mediator.assert_has_calls(expected_calls)

    def test_driver_create_cifs_share_from_snapshot(self):
        self.init_driver()

        expected_location = '\\\\%s\%s' % (constants.EXPECTED_IP_10203040,
                                           constants.EXPECTED_SHARE_NAME)

        self.mock_mediator.create_share_from_snapshot.return_value = (
            constants.EXPECTED_SHARE_NAME)
        hpe3parmediator.HPE3ParMediator = self.real_hpe_3par_mediator

        snapshot_instance = constants.SNAPSHOT_INSTANCE.copy()
        snapshot_instance['protocol'] = constants.CIFS

        location = self.do_create_share_from_snapshot(
            constants.CIFS,
            constants.SHARE_TYPE_ID,
            snapshot_instance,
            constants.EXPECTED_SHARE_ID,
            constants.EXPECTED_SIZE_2)

        self.assertIn(expected_location, location)
        expected_calls = [
            mock.call.create_share_from_snapshot(
                constants.EXPECTED_SHARE_ID,
                constants.CIFS,
                constants.EXPECTED_EXTRA_SPECS,
                constants.EXPECTED_FSTORE,
                constants.EXPECTED_SHARE_ID,
                constants.EXPECTED_SNAP_ID,
                constants.EXPECTED_FPG,
                constants.EXPECTED_VFS,
                [constants.EXPECTED_IP_10203040],
                comment=mock.ANY,
                size=constants.EXPECTED_SIZE_2),
        ]
        self.mock_mediator.assert_has_calls(expected_calls)

    def test_driver_create_nfs_share_from_snapshot(self):
        self.init_driver()

        expected_location = ':'.join((constants.EXPECTED_IP_10203040,
                                      constants.EXPECTED_SHARE_PATH))

        self.mock_mediator.create_share_from_snapshot.return_value = (
            constants.EXPECTED_SHARE_PATH)
        hpe3parmediator.HPE3ParMediator = self.real_hpe_3par_mediator

        location = self.do_create_share_from_snapshot(
            constants.NFS,
            constants.SHARE_TYPE_ID,
            constants.SNAPSHOT_INSTANCE,
            constants.EXPECTED_SHARE_ID,
            constants.EXPECTED_SIZE_1)

        self.assertIn(expected_location, location)
        expected_calls = [
            mock.call.create_share_from_snapshot(
                constants.EXPECTED_SHARE_ID,
                constants.NFS,
                constants.EXPECTED_EXTRA_SPECS,
                constants.EXPECTED_PROJECT_ID,
                constants.EXPECTED_SHARE_ID,
                constants.EXPECTED_SNAP_ID,
                constants.EXPECTED_FPG,
                constants.EXPECTED_VFS,
                [constants.EXPECTED_IP_10203040],
                comment=mock.ANY,
                size=constants.EXPECTED_SIZE_1),
        ]

        self.mock_mediator.assert_has_calls(expected_calls)

    def test_driver_delete_share(self):
        self.init_driver()

        context = None
        share_server = None
        share = {
            'project_id': constants.EXPECTED_PROJECT_ID,
            'id': constants.EXPECTED_SHARE_ID,
            'share_proto': constants.CIFS,
            'size': constants.EXPECTED_SIZE_1,
            'host': constants.EXPECTED_HOST
        }

        self.driver.delete_share(context, share, share_server)

        expected_calls = [
            mock.call.delete_share(constants.EXPECTED_PROJECT_ID,
                                   constants.EXPECTED_SHARE_ID,
                                   constants.EXPECTED_SIZE_1,
                                   constants.CIFS,
                                   constants.EXPECTED_FPG,
                                   constants.EXPECTED_VFS,
                                   constants.EXPECTED_IP_10203040)]

        self.mock_mediator.assert_has_calls(expected_calls)

    def test_driver_create_snapshot(self):
        self.init_driver()

        context = None
        share_server = None
        self.driver.create_snapshot(context,
                                    constants.SNAPSHOT_INFO,
                                    share_server)

        expected_calls = [
            mock.call.create_snapshot(constants.EXPECTED_PROJECT_ID,
                                      constants.EXPECTED_SHARE_ID,
                                      constants.NFS,
                                      constants.EXPECTED_SNAP_ID,
                                      constants.EXPECTED_FPG,
                                      constants.EXPECTED_VFS)]
        self.mock_mediator.assert_has_calls(expected_calls)

    def test_driver_delete_snapshot(self):
        self.init_driver()

        context = None
        share_server = None
        self.driver.delete_snapshot(context,
                                    constants.SNAPSHOT_INFO,
                                    share_server)

        expected_calls = [
            mock.call.delete_snapshot(constants.EXPECTED_PROJECT_ID,
                                      constants.EXPECTED_SHARE_ID,
                                      constants.NFS,
                                      constants.EXPECTED_SNAP_ID,
                                      constants.EXPECTED_FPG,
                                      constants.EXPECTED_VFS)
        ]
        self.mock_mediator.assert_has_calls(expected_calls)

    def test_driver_update_access_add_rule(self):
        self.init_driver()

        context = None

        self.driver.update_access(context,
                                  constants.NFS_SHARE_INFO,
                                  [constants.ACCESS_RULE_NFS],
                                  [constants.ADD_RULE_IP],
                                  [],
                                  constants.SHARE_SERVER)

        expected_calls = [
            mock.call.update_access(constants.EXPECTED_PROJECT_ID,
                                    constants.EXPECTED_SHARE_ID,
                                    constants.NFS,
                                    constants.EXPECTED_EXTRA_SPECS,
                                    [constants.ACCESS_RULE_NFS],
                                    [constants.ADD_RULE_IP],
                                    [],
                                    constants.EXPECTED_FPG,
                                    constants.EXPECTED_VFS)
        ]
        self.mock_mediator.assert_has_calls(expected_calls)

    def test_driver_update_access_delete_rule(self):
        self.init_driver()

        context = None

        self.driver.update_access(context,
                                  constants.NFS_SHARE_INFO,
                                  [constants.ACCESS_RULE_NFS],
                                  [],
                                  [constants.DELETE_RULE_IP],
                                  constants.SHARE_SERVER)

        expected_calls = [
            mock.call.update_access(constants.EXPECTED_PROJECT_ID,
                                    constants.EXPECTED_SHARE_ID,
                                    constants.NFS,
                                    constants.EXPECTED_EXTRA_SPECS,
                                    [constants.ACCESS_RULE_NFS],
                                    [],
                                    [constants.DELETE_RULE_IP],
                                    constants.EXPECTED_FPG,
                                    constants.EXPECTED_VFS)
        ]
        self.mock_mediator.assert_has_calls(expected_calls)

    def test_driver_extend_share(self):
        self.init_driver()

        old_size = constants.NFS_SHARE_INFO['size']
        new_size = old_size * 2

        share_server = None
        self.driver.extend_share(constants.NFS_SHARE_INFO,
                                 new_size, share_server)

        self.mock_mediator.resize_share.assert_called_once_with(
            constants.EXPECTED_PROJECT_ID,
            constants.EXPECTED_SHARE_ID,
            constants.NFS,
            new_size,
            old_size,
            constants.EXPECTED_FPG,
            constants.EXPECTED_VFS)

    def test_driver_shrink_share(self):
        self.init_driver()

        old_size = constants.NFS_SHARE_INFO['size']
        new_size = old_size / 2
        share_server = None
        self.driver.shrink_share(constants.NFS_SHARE_INFO,
                                 new_size, share_server)

        self.mock_mediator.resize_share.assert_called_once_with(
            constants.EXPECTED_PROJECT_ID,
            constants.EXPECTED_SHARE_ID,
            constants.NFS,
            new_size,
            old_size,
            constants.EXPECTED_FPG,
            constants.EXPECTED_VFS)

    def test_driver_get_share_stats_not_ready(self):
        """Protect against stats update before driver is ready."""

        self.mock_object(hpe3pardriver, 'LOG')

        expected_result = {
            'driver_handles_share_servers': True,
            'qos': False,
            'driver_version': self.driver.VERSION,
            'free_capacity_gb': 0,
            'max_over_subscription_ratio': None,
            'reserved_percentage': 0,
            'provisioned_capacity_gb': 0,
            'share_backend_name': 'HPE_3PAR',
            'snapshot_support': True,
            'create_share_from_snapshot_support': True,
            'revert_to_snapshot_support': False,
            'mount_snapshot_support': False,
            'share_group_stats': {
                'consistent_snapshot_support': None,
            },
            'storage_protocol': 'NFS_CIFS',
            'thin_provisioning': True,
            'total_capacity_gb': 0,
            'vendor_name': 'HPE',
            'pools': None,
            'replication_domain': None,
            'filter_function': None,
            'goodness_function': None,
            'ipv4_support': True,
            'ipv6_support': False,
        }

        result = self.driver.get_share_stats(refresh=True)
        self.assertEqual(expected_result, result)

        expected_calls = [
            mock.call.info('Skipping capacity and capabilities update. '
                           'Setup has not completed.')
        ]
        hpe3pardriver.LOG.assert_has_calls(expected_calls)

    def test_driver_get_share_stats_no_refresh(self):
        """Driver does not call mediator when refresh=False."""

        self.init_driver()
        self.driver._stats = constants.EXPECTED_STATS

        result = self.driver.get_share_stats(refresh=False)

        self.assertEqual(constants.EXPECTED_STATS, result)
        self.assertEqual([], self.mock_mediator.mock_calls)

    def test_driver_get_share_stats_with_refresh(self):
        """Driver adds stats from mediator to expected structure."""

        self.init_driver()
        expected_free = constants.EXPECTED_SIZE_1
        expected_capacity = constants.EXPECTED_SIZE_2
        expected_version = self.driver.VERSION

        self.mock_mediator.get_fpg_status.return_value = {
            'pool_name': constants.EXPECTED_FPG,
            'total_capacity_gb': expected_capacity,
            'free_capacity_gb': expected_free,
            'thin_provisioning': True,
            'dedupe': False,
            'hpe3par_flash_cache': False,
            'hp3par_flash_cache': False,
            'reserved_percentage': 0,
            'provisioned_capacity_gb': expected_capacity
        }

        expected_result = {
            'share_backend_name': 'HPE_3PAR',
            'vendor_name': 'HPE',
            'driver_version': expected_version,
            'storage_protocol': 'NFS_CIFS',
            'driver_handles_share_servers': True,
            'total_capacity_gb': 0,
            'free_capacity_gb': 0,
            'provisioned_capacity_gb': 0,
            'reserved_percentage': 0,
            'max_over_subscription_ratio': None,
            'qos': False,
            'thin_provisioning': True,
            'pools': [{
                'pool_name': constants.EXPECTED_FPG,
                'total_capacity_gb': expected_capacity,
                'free_capacity_gb': expected_free,
                'thin_provisioning': True,
                'dedupe': False,
                'hpe3par_flash_cache': False,
                'hp3par_flash_cache': False,
                'reserved_percentage': 0,
                'provisioned_capacity_gb': expected_capacity}],
            'snapshot_support': True,
            'create_share_from_snapshot_support': True,
            'revert_to_snapshot_support': False,
            'mount_snapshot_support': False,
            'share_group_stats': {
                'consistent_snapshot_support': None,
            },
            'replication_domain': None,
            'filter_function': None,
            'goodness_function': None,
            'ipv4_support': True,
            'ipv6_support': False,
        }

        result = self.driver.get_share_stats(refresh=True)
        self.assertEqual(expected_result, result)

        expected_calls = [
            mock.call.get_fpg_status(constants.EXPECTED_FPG)
        ]
        self.mock_mediator.assert_has_calls(expected_calls)
        self.assertTrue(self.mock_mediator.get_fpg_status.called)

    def test_driver_get_share_stats_premature(self):
        """Driver init stats before init_driver completed."""

        expected_version = self.driver.VERSION

        self.mock_mediator.get_fpg_status.return_value = {'not_called': 1}

        expected_result = {
            'qos': False,
            'driver_handles_share_servers': True,
            'driver_version': expected_version,
            'free_capacity_gb': 0,
            'max_over_subscription_ratio': None,
            'pools': None,
            'provisioned_capacity_gb': 0,
            'reserved_percentage': 0,
            'share_backend_name': 'HPE_3PAR',
            'storage_protocol': 'NFS_CIFS',
            'thin_provisioning': True,
            'total_capacity_gb': 0,
            'vendor_name': 'HPE',
            'snapshot_support': True,
            'create_share_from_snapshot_support': True,
            'revert_to_snapshot_support': False,
            'mount_snapshot_support': False,
            'share_group_stats': {
                'consistent_snapshot_support': None,
            },
            'replication_domain': None,
            'filter_function': None,
            'goodness_function': None,
            'ipv4_support': True,
            'ipv6_support': False,
        }

        result = self.driver.get_share_stats(refresh=True)
        self.assertEqual(expected_result, result)
        self.assertFalse(self.mock_mediator.get_fpg_status.called)

    @ddt.data(('test"dquote', 'test_dquote'),
              ("test'squote", "test_squote"),
              ('test-:;,.punc', 'test-:_punc'),
              ('test with spaces ', 'test with spaces '),
              ('x' * 300, 'x' * 300))
    @ddt.unpack
    def test_build_comment(self, display_name, clean_name):

        host = 'test-stack1@backend#pool'
        share = {
            'host': host,
            'display_name': display_name
        }
        comment = self.driver.build_share_comment(share)

        cleaned = {
            'host': host,
            'clean_name': clean_name
        }

        expected = ("OpenStack Manila - host=%(host)s  "
                    "orig_name=%(clean_name)s created=" % cleaned)[:254]

        self.assertLess(len(comment), 255)
        self.assertTrue(comment.startswith(expected))

        # Test for some chars that are not allowed.
        # Don't test with same regex as the code uses.
        for c in "'\".,;":
            self.assertNotIn(c, comment)

    def test_get_network_allocations_number(self):
        self.assertEqual(1, self.driver.get_network_allocations_number())

    def test_setup_server(self):
        """Setup server by creating a new FSIP."""

        self.init_driver()

        network_info = {
            'network_allocations': [
                {'ip_address': constants.EXPECTED_IP_1234}],
            'cidr': '/'.join((constants.EXPECTED_IP_1234,
                              constants.CIDR_PREFIX)),
            'network_type': constants.EXPECTED_VLAN_TYPE,
            'segmentation_id': constants.EXPECTED_VLAN_TAG,
            'server_id': constants.EXPECTED_SERVER_ID,
        }

        expected_result = {
            'share_server_name': constants.EXPECTED_SERVER_ID,
            'share_server_id': constants.EXPECTED_SERVER_ID,
            'ip': constants.EXPECTED_IP_1234,
            'subnet': constants.EXPECTED_SUBNET,
            'vlantag': constants.EXPECTED_VLAN_TAG,
            'fpg': constants.EXPECTED_FPG,
            'vfs': constants.EXPECTED_VFS,
        }
        metadata = {'request_host': constants.EXPECTED_HOST}
        result = self.driver._setup_server(network_info, metadata)

        expected_calls = [
            mock.call.create_fsip(constants.EXPECTED_IP_1234,
                                  constants.EXPECTED_SUBNET,
                                  constants.EXPECTED_VLAN_TAG,
                                  constants.EXPECTED_FPG,
                                  constants.EXPECTED_VFS)
        ]
        self.mock_mediator.assert_has_calls(expected_calls)

        self.assertEqual(expected_result, result)

    def test_setup_server_fails_for_unsupported_network_type(self):
        """Setup server fails for unsupported network type"""

        self.init_driver()

        network_info = {
            'network_allocations': [
                {'ip_address': constants.EXPECTED_IP_1234}],
            'cidr': '/'.join((constants.EXPECTED_IP_1234,
                              constants.CIDR_PREFIX)),
            'network_type': constants.EXPECTED_VXLAN_TYPE,
            'segmentation_id': constants.EXPECTED_VLAN_TAG,
            'server_id': constants.EXPECTED_SERVER_ID,
        }
        metadata = {'request_host': constants.EXPECTED_HOST}

        self.assertRaises(exception.NetworkBadConfigurationException,
                          self.driver._setup_server,
                          network_info, metadata)

    def test_setup_server_fails_for_exceed_pool_max_supported_ips(self):
        """Setup server fails when the VFS has reached max supported IPs"""

        self.init_driver()

        network_info = {
            'network_allocations': [
                {'ip_address': constants.EXPECTED_IP_1234}],
            'cidr': '/'.join((constants.EXPECTED_IP_1234,
                              constants.CIDR_PREFIX)),
            'network_type': constants.EXPECTED_VLAN_TYPE,
            'segmentation_id': constants.EXPECTED_VLAN_TAG,
            'server_id': constants.EXPECTED_SERVER_ID,
        }
        metadata = {'request_host': constants.EXPECTED_HOST}

        expected_vfs = self.driver.fpgs[
            constants.EXPECTED_FPG][constants.EXPECTED_VFS]
        self.driver.fpgs[constants.EXPECTED_FPG][constants.EXPECTED_VFS] = [
            '10.0.0.1', '10.0.0.2', '10.0.0.3', '10.0.0.4']

        self.assertRaises(exception.Invalid,
                          self.driver._setup_server,
                          network_info, metadata)
        self.driver.fpgs[constants.EXPECTED_FPG][constants.EXPECTED_VFS
                                                 ] = expected_vfs

    def test_teardown_server(self):
        """Test tear down server"""

        self.init_driver()

        server_details = {
            'ip': constants.EXPECTED_IP_10203040,
            'fpg': constants.EXPECTED_FPG,
            'vfs': constants.EXPECTED_VFS,
        }

        self.driver._teardown_server(server_details)

        expected_calls = [
            mock.call.remove_fsip(constants.EXPECTED_IP_10203040,
                                  constants.EXPECTED_FPG,
                                  constants.EXPECTED_VFS)
        ]
        self.mock_mediator.assert_has_calls(expected_calls)



# Copyright (c) 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import socket

import ddt
import mock
from oslo_config import cfg

from manila import context
from manila import exception
from manila.share import configuration as config
from manila.share.drivers import ganesha
from manila.share.drivers import glusterfs
from manila.share.drivers.glusterfs import layout
from manila import test
from manila.tests import fake_share
from manila.tests import fake_utils


CONF = cfg.CONF


fake_gluster_manager_attrs = {
    'export': '127.0.0.1:/testvol',
    'host': '127.0.0.1',
    'qualified': 'testuser@127.0.0.1:/testvol',
    'user': 'testuser',
    'volume': 'testvol',
    'path_to_private_key': '/fakepath/to/privatekey',
    'remote_server_password': 'fakepassword',
}

fake_share_name = 'fakename'
NFS_EXPORT_DIR = 'nfs.export-dir'
NFS_EXPORT_VOL = 'nfs.export-volumes'
NFS_RPC_AUTH_ALLOW = 'nfs.rpc-auth-allow'
NFS_RPC_AUTH_REJECT = 'nfs.rpc-auth-reject'


@ddt.ddt
class GlusterfsShareDriverTestCase(test.TestCase):
    """Tests GlusterfsShareDriver."""

    def setUp(self):
        super(GlusterfsShareDriverTestCase, self).setUp()
        fake_utils.stub_out_utils_execute(self)
        self._execute = fake_utils.fake_execute
        self._context = context.get_admin_context()
        self.addCleanup(fake_utils.fake_execute_set_repliers, [])
        self.addCleanup(fake_utils.fake_execute_clear_log)

        CONF.set_default('reserved_share_percentage', 50)
        CONF.set_default('driver_handles_share_servers', False)

        self.fake_conf = config.Configuration(None)
        self._driver = glusterfs.GlusterfsShareDriver(
            execute=self._execute,
            configuration=self.fake_conf)
        self.share = fake_share.fake_share(share_proto='NFS')

    def test_do_setup(self):
        self.mock_object(self._driver, '_get_helper')
        self.mock_object(layout.GlusterfsShareDriverBase, 'do_setup')
        _context = mock.Mock()

        self._driver.do_setup(_context)

        self._driver._get_helper.assert_called_once_with()
        layout.GlusterfsShareDriverBase.do_setup.assert_called_once_with(
            _context)

    @ddt.data(True, False)
    def test_setup_via_manager(self, has_parent):
        gmgr = mock.Mock()
        share_mgr_parent = mock.Mock() if has_parent else None
        nfs_helper = mock.Mock()
        nfs_helper.get_export = mock.Mock(return_value='host:/vol')
        self._driver.nfs_helper = mock.Mock(return_value=nfs_helper)

        ret = self._driver._setup_via_manager(
            {'manager': gmgr, 'share': self.share},
            share_manager_parent=share_mgr_parent)

        gmgr.set_vol_option.assert_called_once_with(
            'nfs.export-volumes', False)
        self._driver.nfs_helper.assert_called_once_with(
            self._execute, self.fake_conf, gluster_manager=gmgr)
        nfs_helper.get_export.assert_called_once_with(self.share)
        self.assertEqual('host:/vol', ret)

    @ddt.data({'helpercls': None, 'path': '/fakepath'},
              {'helpercls': None, 'path': None},
              {'helpercls': glusterfs.GlusterNFSHelper, 'path': '/fakepath'},
              {'helpercls': glusterfs.GlusterNFSHelper, 'path': None})
    @ddt.unpack
    def test_setup_via_manager_path(self, helpercls, path):
        gmgr = mock.Mock()
        gmgr.path = path
        if not helpercls:
            helper = mock.Mock()
            helper.get_export = mock.Mock(return_value='host:/vol')
            helpercls = mock.Mock(return_value=helper)
        self._driver.nfs_helper = helpercls
        if helpercls == glusterfs.GlusterNFSHelper and path is None:
            gmgr.get_vol_option = mock.Mock(return_value=True)

        self._driver._setup_via_manager(
            {'manager': gmgr, 'share': self.share})

        if helpercls == glusterfs.GlusterNFSHelper and path is None:
            gmgr.get_vol_option.assert_called_once_with(
                NFS_EXPORT_VOL, boolean=True)
            args = (NFS_RPC_AUTH_REJECT, '*')
        else:
            args = (NFS_EXPORT_VOL, False)
        gmgr.set_vol_option.assert_called_once_with(*args)

    def test_setup_via_manager_export_volumes_off(self):
        gmgr = mock.Mock()
        gmgr.path = None
        gmgr.get_vol_option = mock.Mock(return_value=False)
        self._driver.nfs_helper = glusterfs.GlusterNFSHelper

        self.assertRaises(exception.GlusterfsException,
                          self._driver._setup_via_manager,
                          {'manager': gmgr, 'share': self.share})

        gmgr.get_vol_option.assert_called_once_with(NFS_EXPORT_VOL,
                                                    boolean=True)

    def test_check_for_setup_error(self):
        self._driver.check_for_setup_error()

    def test_update_share_stats(self):
        self.mock_object(layout.GlusterfsShareDriverBase,
                         '_update_share_stats')

        self._driver._update_share_stats()

        (layout.GlusterfsShareDriverBase._update_share_stats.
         assert_called_once_with({'storage_protocol': 'NFS',
                                  'vendor_name': 'Red Hat',
                                  'share_backend_name': 'GlusterFS',
                                  'reserved_percentage': 50}))

    def test_get_network_allocations_number(self):
        self.assertEqual(0, self._driver.get_network_allocations_number())

    def test_get_helper(self):
        ret = self._driver._get_helper()
        self.assertIsInstance(ret, self._driver.nfs_helper)

    @ddt.data({'path': '/fakepath', 'helper': glusterfs.GlusterNFSHelper},
              {'path': None, 'helper': glusterfs.GlusterNFSVolHelper})
    @ddt.unpack
    def test_get_helper_vol(self, path, helper):
        self._driver.nfs_helper = glusterfs.GlusterNFSHelper

        gmgr = mock.Mock(path=path)
        ret = self._driver._get_helper(gmgr)

        self.assertIsInstance(ret, helper)

    @ddt.data('type', 'level')
    def test_supported_access_features(self, feature):
        nfs_helper = mock.Mock()
        supported_access_feature = mock.Mock()
        setattr(nfs_helper, 'supported_access_%ss' % feature,
                supported_access_feature)
        self.mock_object(self._driver, 'nfs_helper', nfs_helper)

        ret = getattr(self._driver, 'supported_access_%ss' % feature)

        self.assertEqual(supported_access_feature, ret)

    def test_update_access_via_manager(self):
        self.mock_object(self._driver, '_get_helper')
        gmgr = mock.Mock()
        add_rules = mock.Mock()
        delete_rules = mock.Mock()

        self._driver._update_access_via_manager(
            gmgr, self._context, self.share,
            add_rules, delete_rules, recovery=True)

        self._driver._get_helper.assert_called_once_with(gmgr)
        self._driver._get_helper().update_access.assert_called_once_with(
            '/', self.share, add_rules, delete_rules, recovery=True)


@ddt.ddt
class GlusterNFSHelperTestCase(test.TestCase):
    """Tests GlusterNFSHelper."""

    def setUp(self):
        super(GlusterNFSHelperTestCase, self).setUp()
        fake_utils.stub_out_utils_execute(self)
        gluster_manager = mock.Mock(**fake_gluster_manager_attrs)
        self._execute = mock.Mock(return_value=('', ''))
        self.fake_conf = config.Configuration(None)
        self._helper = glusterfs.GlusterNFSHelper(
            self._execute, self.fake_conf, gluster_manager=gluster_manager)

    def test_get_export(self):
        ret = self._helper.get_export(mock.Mock())

        self.assertEqual(fake_gluster_manager_attrs['export'], ret)

    @ddt.data({'output_str': '/foo(10.0.0.1|10.0.0.2),/bar(10.0.0.1)',
               'expected': {'foo': ['10.0.0.1', '10.0.0.2'],
                            'bar': ['10.0.0.1']}},
              {'output_str': None, 'expected': {}})
    @ddt.unpack
    def test_get_export_dir_dict(self, output_str, expected):
        self.mock_object(self._helper.gluster_manager,
                         'get_vol_option',
                         mock.Mock(return_value=output_str))

        ret = self._helper._get_export_dir_dict()

        self.assertEqual(expected, ret)
        (self._helper.gluster_manager.get_vol_option.
         assert_called_once_with(NFS_EXPORT_DIR))

    @ddt.data({'delta': (['10.0.0.2'], []), 'extra_exports': {},
               'new_exports': '/fakename(10.0.0.1|10.0.0.2)'},
              {'delta': (['10.0.0.1'], []), 'extra_exports': {},
               'new_exports': '/fakename(10.0.0.1)'},
              {'delta': ([], ['10.0.0.2']), 'extra_exports': {},
               'new_exports': '/fakename(10.0.0.1)'},
              {'delta': ([], ['10.0.0.1']), 'extra_exports': {},
               'new_exports': None},
              {'delta': ([], ['10.0.0.1']),
               'extra_exports': {'elsewhere': ['10.0.1.3']},
               'new_exports': '/elsewhere(10.0.1.3)'})
    @ddt.unpack
    def test_update_access(self, delta, extra_exports, new_exports):
        gluster_manager_attrs = {'path': '/fakename'}
        gluster_manager_attrs.update(fake_gluster_manager_attrs)
        gluster_mgr = mock.Mock(**gluster_manager_attrs)
        helper = glusterfs.GlusterNFSHelper(
            self._execute, self.fake_conf, gluster_manager=gluster_mgr)
        export_dir_dict = {'fakename': ['10.0.0.1']}
        export_dir_dict.update(extra_exports)
        helper._get_export_dir_dict = mock.Mock(return_value=export_dir_dict)
        _share = mock.Mock()

        add_rules, delete_rules = (
            map(lambda a: {'access_to': a}, r) for r in delta)
        helper.update_access('/', _share, add_rules, delete_rules)

        helper._get_export_dir_dict.assert_called_once_with()
        gluster_mgr.set_vol_option.assert_called_once_with(NFS_EXPORT_DIR,
                                                           new_exports)

    @ddt.data({}, {'elsewhere': '10.0.1.3'})
    def test_update_access_disjoint(self, export_dir_dict):
        gluster_manager_attrs = {'path': '/fakename'}
        gluster_manager_attrs.update(fake_gluster_manager_attrs)
        gluster_mgr = mock.Mock(**gluster_manager_attrs)
        helper = glusterfs.GlusterNFSHelper(
            self._execute, self.fake_conf, gluster_manager=gluster_mgr)
        helper._get_export_dir_dict = mock.Mock(return_value=export_dir_dict)
        _share = mock.Mock()

        helper.update_access('/', _share, [], [{'access_to': '10.0.0.2'}])

        helper._get_export_dir_dict.assert_called_once_with()
        self.assertFalse(gluster_mgr.set_vol_option.called)


@ddt.ddt
class GlusterNFSVolHelperTestCase(test.TestCase):
    """Tests GlusterNFSVolHelper."""

    def setUp(self):
        super(GlusterNFSVolHelperTestCase, self).setUp()
        fake_utils.stub_out_utils_execute(self)
        gluster_manager = mock.Mock(**fake_gluster_manager_attrs)
        self._execute = mock.Mock(return_value=('', ''))
        self.fake_conf = config.Configuration(None)
        self._helper = glusterfs.GlusterNFSVolHelper(
            self._execute, self.fake_conf, gluster_manager=gluster_manager)

    @ddt.data({'output_str': '10.0.0.1,10.0.0.2',
               'expected': ['10.0.0.1', '10.0.0.2']},
              {'output_str': None, 'expected': []})
    @ddt.unpack
    def test_get_vol_exports(self, output_str, expected):
        self.mock_object(self._helper.gluster_manager,
                         'get_vol_option',
                         mock.Mock(return_value=output_str))

        ret = self._helper._get_vol_exports()

        self.assertEqual(expected, ret)
        (self._helper.gluster_manager.get_vol_option.
         assert_called_once_with(NFS_RPC_AUTH_ALLOW))

    @ddt.data({'delta': (["10.0.0.1"], []), 'expected': "10.0.0.1,10.0.0.3"},
              {'delta': (["10.0.0.2"], []),
               'expected': "10.0.0.1,10.0.0.2,10.0.0.3"},
              {'delta': ([], ["10.0.0.1"]), 'expected': "10.0.0.3"},
              {'delta': ([], ["10.0.0.2"]), 'expected': "10.0.0.1,10.0.0.3"})
    @ddt.unpack
    def test_update_access(self, delta, expected):
        self.mock_object(self._helper, '_get_vol_exports', mock.Mock(
            return_value=["10.0.0.1", "10.0.0.3"]))
        _share = mock.Mock()

        add_rules, delete_rules = (
            map(lambda a: {'access_to': a}, r) for r in delta)
        self._helper.update_access("/", _share, add_rules, delete_rules)

        self._helper._get_vol_exports.assert_called_once_with()
        argseq = [(NFS_RPC_AUTH_ALLOW, expected), (NFS_RPC_AUTH_REJECT, None)]
        self.assertEqual(
            [mock.call(*a) for a in argseq],
            self._helper.gluster_manager.set_vol_option.call_args_list)

    def test_update_access_empty(self):
        self.mock_object(self._helper, '_get_vol_exports', mock.Mock(
            return_value=["10.0.0.1"]))
        _share = mock.Mock()

        self._helper.update_access("/", _share, [],
                                   [{'access_to': "10.0.0.1"}])

        self._helper._get_vol_exports.assert_called_once_with()
        argseq = [(NFS_RPC_AUTH_ALLOW, None), (NFS_RPC_AUTH_REJECT, "*")]
        self.assertEqual(
            [mock.call(*a) for a in argseq],
            self._helper.gluster_manager.set_vol_option.call_args_list)


class GaneshaNFSHelperTestCase(test.TestCase):
    """Tests GaneshaNFSHelper."""

    def setUp(self):
        super(GaneshaNFSHelperTestCase, self).setUp()
        self.gluster_manager = mock.Mock(**fake_gluster_manager_attrs)
        self._execute = mock.Mock(return_value=('', ''))
        self._root_execute = mock.Mock(return_value=('', ''))
        self.access = fake_share.fake_access()
        self.fake_conf = config.Configuration(None)
        self.fake_template = {'key': 'value'}
        self.share = fake_share.fake_share()
        self.mock_object(glusterfs.ganesha_utils, 'RootExecutor',
                         mock.Mock(return_value=self._root_execute))
        self.mock_object(glusterfs.ganesha.GaneshaNASHelper, '__init__',
                         mock.Mock())
        socket.gethostname = mock.Mock(return_value='example.com')
        self._helper = glusterfs.GaneshaNFSHelper(
            self._execute, self.fake_conf,
            gluster_manager=self.gluster_manager)
        self._helper.tag = 'GLUSTER-Ganesha-localhost'

    def test_init_local_ganesha_server(self):
        glusterfs.ganesha_utils.RootExecutor.assert_called_once_with(
            self._execute)
        socket.gethostname.assert_has_calls([mock.call()])
        glusterfs.ganesha.GaneshaNASHelper.__init__.assert_has_calls(
            [mock.call(self._root_execute, self.fake_conf,
                       tag='GLUSTER-Ganesha-example.com')])

    def test_get_export(self):
        ret = self._helper.get_export(self.share)

        self.assertEqual('example.com:/fakename--<access-id>', ret)

    def test_init_remote_ganesha_server(self):
        ssh_execute = mock.Mock(return_value=('', ''))
        CONF.set_default('glusterfs_ganesha_server_ip', 'fakeip')
        self.mock_object(glusterfs.ganesha_utils, 'SSHExecutor',
                         mock.Mock(return_value=ssh_execute))
        glusterfs.GaneshaNFSHelper(
            self._execute, self.fake_conf,
            gluster_manager=self.gluster_manager)
        glusterfs.ganesha_utils.SSHExecutor.assert_called_once_with(
            'fakeip', 22, None, 'root', password=None, privatekey=None)
        glusterfs.ganesha.GaneshaNASHelper.__init__.assert_has_calls(
            [mock.call(ssh_execute, self.fake_conf,
                       tag='GLUSTER-Ganesha-fakeip')])

    def test_init_helper(self):
        ganeshelper = mock.Mock()
        exptemp = mock.Mock()

        def set_attributes(*a, **kw):
            self._helper.ganesha = ganeshelper
            self._helper.export_template = exptemp

        self.mock_object(ganesha.GaneshaNASHelper, 'init_helper',
                         mock.Mock(side_effect=set_attributes))
        self.assertEqual({}, glusterfs.GaneshaNFSHelper.shared_data)

        self._helper.init_helper()

        ganesha.GaneshaNASHelper.init_helper.assert_called_once_with()
        self.assertEqual(ganeshelper, self._helper.ganesha)
        self.assertEqual(exptemp, self._helper.export_template)
        self.assertEqual({
            'GLUSTER-Ganesha-localhost': {
                'ganesha': ganeshelper,
                'export_template': exptemp}},
            glusterfs.GaneshaNFSHelper.shared_data)

        other_helper = glusterfs.GaneshaNFSHelper(
            self._execute, self.fake_conf,
            gluster_manager=self.gluster_manager)
        other_helper.tag = 'GLUSTER-Ganesha-localhost'

        other_helper.init_helper()

        self.assertEqual(ganeshelper, other_helper.ganesha)
        self.assertEqual(exptemp, other_helper.export_template)

    def test_default_config_hook(self):
        fake_conf_dict = {'key': 'value1'}
        mock_ganesha_utils_patch = mock.Mock()

        def fake_patch_run(tmpl1, tmpl2):
            mock_ganesha_utils_patch(
                copy.deepcopy(tmpl1), tmpl2)
            tmpl1.update(tmpl2)

        self.mock_object(glusterfs.ganesha.GaneshaNASHelper,
                         '_default_config_hook',
                         mock.Mock(return_value=self.fake_template))
        self.mock_object(glusterfs.ganesha_utils, 'path_from',
                         mock.Mock(return_value='/fakedir/glusterfs/conf'))
        self.mock_object(self._helper, '_load_conf_dir',
                         mock.Mock(return_value=fake_conf_dict))
        self.mock_object(glusterfs.ganesha_utils, 'patch',
                         mock.Mock(side_effect=fake_patch_run))

        ret = self._helper._default_config_hook()

        (glusterfs.ganesha.GaneshaNASHelper._default_config_hook.
            assert_called_once_with())
        glusterfs.ganesha_utils.path_from.assert_called_once_with(
            glusterfs.__file__, 'conf')
        self._helper._load_conf_dir.assert_called_once_with(
            '/fakedir/glusterfs/conf')
        glusterfs.ganesha_utils.patch.assert_called_once_with(
            self.fake_template, fake_conf_dict)
        self.assertEqual(fake_conf_dict, ret)

    def test_fsal_hook(self):
        self._helper.gluster_manager.path = '/fakename'
        output = {
            'Hostname': '127.0.0.1',
            'Volume': 'testvol',
            'Volpath': '/fakename'
        }

        ret = self._helper._fsal_hook('/fakepath', self.share, self.access)

        self.assertEqual(output, ret)



# Copyright 2011 OpenStack LLC.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""RequestContext: context for requests that persist through all of manila."""

import copy

from oslo_context import context
from oslo_utils import timeutils
import six

from manila.i18n import _
from manila import policy


class RequestContext(context.RequestContext):
    """Security context and request information.

    Represents the user taking a given action within the system.

    """

    def __init__(self, user_id, project_id, is_admin=None, read_deleted="no",
                 roles=None, remote_address=None, timestamp=None,
                 request_id=None, auth_token=None, overwrite=True,
                 quota_class=None, service_catalog=None, **kwargs):
        """Initialize RequestContext.

        :param read_deleted: 'no' indicates deleted records are hidden, 'yes'
            indicates deleted records are visible, 'only' indicates that
            *only* deleted records are visible.

        :param overwrite: Set to False to ensure that the greenthread local
            copy of the index is not overwritten.

        :param kwargs: Extra arguments that might be present, but we ignore
            because they possibly came in from older rpc messages.
        """

        user = kwargs.pop('user', None)
        tenant = kwargs.pop('tenant', None)
        super(RequestContext, self).__init__(
            auth_token=auth_token,
            user=user_id or user,
            tenant=project_id or tenant,
            domain=kwargs.pop('domain', None),
            user_domain=kwargs.pop('user_domain', None),
            project_domain=kwargs.pop('project_domain', None),
            is_admin=is_admin,
            read_only=kwargs.pop('read_only', False),
            show_deleted=kwargs.pop('show_deleted', False),
            request_id=request_id,
            resource_uuid=kwargs.pop('resource_uuid', None),
            overwrite=overwrite,
            roles=roles)

        self.user_id = self.user
        self.project_id = self.tenant

        if self.is_admin is None:
            self.is_admin = policy.check_is_admin(self)
        elif self.is_admin and 'admin' not in self.roles:
            self.roles.append('admin')
        self.read_deleted = read_deleted
        self.remote_address = remote_address
        if not timestamp:
            timestamp = timeutils.utcnow()
        if isinstance(timestamp, six.string_types):
            timestamp = timeutils.parse_strtime(timestamp)
        self.timestamp = timestamp
        if service_catalog:
            self.service_catalog = [s for s in service_catalog
                                    if s.get('type') in ('compute', 'volume')]
        else:
            self.service_catalog = []

        self.quota_class = quota_class

    def _get_read_deleted(self):
        return self._read_deleted

    def _set_read_deleted(self, read_deleted):
        if read_deleted not in ('no', 'yes', 'only'):
            raise ValueError(_("read_deleted can only be one of 'no', "
                               "'yes' or 'only', not %r") % read_deleted)
        self._read_deleted = read_deleted

    def _del_read_deleted(self):
        del self._read_deleted

    read_deleted = property(_get_read_deleted, _set_read_deleted,
                            _del_read_deleted)

    def to_dict(self):
        values = super(RequestContext, self).to_dict()
        values.update({
            'user_id': getattr(self, 'user_id', None),
            'project_id': getattr(self, 'project_id', None),
            'read_deleted': getattr(self, 'read_deleted', None),
            'remote_address': getattr(self, 'remote_address', None),
            'timestamp': self.timestamp.isoformat() if hasattr(
                self, 'timestamp') else None,
            'quota_class': getattr(self, 'quota_class', None),
            'service_catalog': getattr(self, 'service_catalog', None)})
        return values

    @classmethod
    def from_dict(cls, values):
        return cls(**values)

    def elevated(self, read_deleted=None, overwrite=False):
        """Return a version of this context with admin flag set."""
        ctx = copy.deepcopy(self)
        ctx.is_admin = True

        if 'admin' not in ctx.roles:
            ctx.roles.append('admin')

        if read_deleted is not None:
            ctx.read_deleted = read_deleted

        return ctx

    def to_policy_values(self):
        policy = super(RequestContext, self).to_policy_values()
        policy['is_admin'] = self.is_admin
        return policy


def get_admin_context(read_deleted="no"):
    return RequestContext(user_id=None,
                          project_id=None,
                          is_admin=True,
                          read_deleted=read_deleted,
                          overwrite=False)



# Copyright (c) 2014 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Huawei Nas Driver for Huawei storage arrays."""
from xml.etree import ElementTree as ET

from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils

from manila import exception
from manila.i18n import _
from manila.share import driver


HUAWEI_UNIFIED_DRIVER_REGISTRY = {
    'V3': 'manila.share.drivers.huawei.v3.connection.V3StorageConnection', }


huawei_opts = [
    cfg.StrOpt('manila_huawei_conf_file',
               default='/etc/manila/manila_huawei_conf.xml',
               help='The configuration file for the Manila Huawei driver.')]

CONF = cfg.CONF
CONF.register_opts(huawei_opts)
LOG = log.getLogger(__name__)


class HuaweiNasDriver(driver.ShareDriver):
    """Huawei Share Driver.

    Executes commands relating to Shares.
    Driver version history::

        1.0 - Initial version.
        1.1 - Add shrink share.
              Add extend share.
              Add manage share.
              Add share level(ro).
              Add smartx capabilities.
              Support multi pools in one backend.
        1.2 - Add share server support.
              Add ensure share.
              Add QoS support.
              Add create share from snapshot.
        1.3 - Add manage snapshot.
              Support reporting disk type of pool.
              Add replication support.
    """

    def __init__(self, *args, **kwargs):
        """Do initialization."""
        LOG.debug("Enter into init function of Huawei Driver.")
        super(HuaweiNasDriver, self).__init__((True, False), *args, **kwargs)

        if not self.configuration:
            raise exception.InvalidInput(reason=_(
                "Huawei driver configuration missing."))

        self.configuration.append_config_values(huawei_opts)
        kwargs.pop('configuration')
        self.plugin = importutils.import_object(self.get_backend_driver(),
                                                self.configuration,
                                                **kwargs)

    def check_for_setup_error(self):
        """Returns an error if prerequisites aren't met."""
        self.plugin.check_conf_file()
        self.plugin.check_service()

    def get_backend_driver(self):
        filename = self.configuration.manila_huawei_conf_file

        try:
            tree = ET.parse(filename)
            root = tree.getroot()
        except Exception as err:
            message = (_('Read Huawei config file(%(filename)s)'
                         ' for Manila error: %(err)s')
                       % {'filename': filename,
                          'err': err})
            LOG.error(message)
            raise exception.InvalidInput(reason=message)
        product = root.findtext('Storage/Product')
        backend_driver = HUAWEI_UNIFIED_DRIVER_REGISTRY.get(product)
        if backend_driver is None:
            raise exception.InvalidInput(
                reason=_('Product %s is not supported. Product '
                         'must be set to V3.') % product)

        return backend_driver

    def do_setup(self, context):
        """Any initialization the huawei nas driver does while starting."""
        LOG.debug("Do setup the plugin.")
        self.plugin.connect()

    def create_share(self, context, share, share_server=None):
        """Create a share."""
        LOG.debug("Create a share.")
        location = self.plugin.create_share(share, share_server)
        return location

    def extend_share(self, share, new_size, share_server=None):
        LOG.debug("Extend a share.")
        self.plugin.extend_share(share, new_size, share_server)

    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None):
        """Create a share from snapshot."""
        LOG.debug("Create a share from snapshot %s.", snapshot['snapshot_id'])
        location = self.plugin.create_share_from_snapshot(share, snapshot)
        return location

    def shrink_share(self, share, new_size, share_server=None):
        """Shrinks size of existing share."""
        LOG.debug("Shrink a share.")
        self.plugin.shrink_share(share, new_size, share_server)

    def delete_share(self, context, share, share_server=None):
        """Delete a share."""
        LOG.debug("Delete a share.")
        self.plugin.delete_share(share, share_server)

    def create_snapshot(self, context, snapshot, share_server=None):
        """Create a snapshot."""
        LOG.debug("Create a snapshot.")
        snapshot_name = self.plugin.create_snapshot(snapshot, share_server)
        return {'provider_location': snapshot_name}

    def delete_snapshot(self, context, snapshot, share_server=None):
        """Delete a snapshot."""
        LOG.debug("Delete a snapshot.")
        self.plugin.delete_snapshot(snapshot, share_server)

    def ensure_share(self, context, share, share_server=None):
        """Ensure that share is exported."""
        LOG.debug("Ensure share.")
        location = self.plugin.ensure_share(share, share_server)
        return location

    def allow_access(self, context, share, access, share_server=None):
        """Allow access to the share."""
        LOG.debug("Allow access.")
        self.plugin.allow_access(share, access, share_server)

    def deny_access(self, context, share, access, share_server=None):
        """Deny access to the share."""
        LOG.debug("Deny access.")
        self.plugin.deny_access(share, access, share_server)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules list."""
        LOG.debug("Update access.")
        self.plugin.update_access(share, access_rules,
                                  add_rules, delete_rules, share_server)

    def get_pool(self, share):
        """Return pool name where the share resides on."""
        LOG.debug("Get pool.")
        return self.plugin.get_pool(share)

    def get_network_allocations_number(self):
        """Get number of network interfaces to be created."""
        LOG.debug("Get network allocations number.")
        return self.plugin.get_network_allocations_number()

    def manage_existing(self, share, driver_options):
        """Manage existing share."""
        LOG.debug("Manage existing share to manila.")
        share_size, location = self.plugin.manage_existing(share,
                                                           driver_options)
        return {'size': share_size, 'export_locations': location}

    def manage_existing_snapshot(self, snapshot, driver_options):
        """Manage existing snapshot."""
        LOG.debug("Manage existing snapshot to manila.")
        snapshot_name = self.plugin.manage_existing_snapshot(snapshot,
                                                             driver_options)
        return {'provider_location': snapshot_name}

    def _update_share_stats(self):
        """Retrieve status info from share group."""

        backend_name = self.configuration.safe_get('share_backend_name')
        data = dict(
            share_backend_name=backend_name or 'HUAWEI_NAS_Driver',
            vendor_name='Huawei',
            driver_version='1.3',
            storage_protocol='NFS_CIFS',
            qos=True,
            total_capacity_gb=0.0,
            free_capacity_gb=0.0,
            snapshot_support=self.plugin.snapshot_support,
            create_share_from_snapshot_support=self.plugin.snapshot_support,
            revert_to_snapshot_support=self.plugin.snapshot_support,
        )

        # huawei array doesn't support snapshot replication, so driver can't
        # create replicated snapshot, this's not fit the requirement of
        # replication feature.
        # to avoid this problem, we specify huawei driver can't support
        # snapshot and replication both, as a workaround.
        if not data['snapshot_support'] and self.plugin.replication_support:
            data['replication_type'] = 'dr'

        self.plugin.update_share_stats(data)
        super(HuaweiNasDriver, self)._update_share_stats(data)

    def _setup_server(self, network_info, metadata=None):
        """Set up share server with given network parameters."""
        return self.plugin.setup_server(network_info, metadata)

    def _teardown_server(self, server_details, security_services=None):
        """Teardown share server."""
        return self.plugin.teardown_server(server_details, security_services)

    def create_replica(self, context, replica_list, new_replica,
                       access_rules, replica_snapshots, share_server=None):
        """Replicate the active replica to a new replica on this backend."""
        return self.plugin.create_replica(context,
                                          replica_list,
                                          new_replica,
                                          access_rules,
                                          replica_snapshots,
                                          share_server)

    def update_replica_state(self, context, replica_list, replica,
                             access_rules, replica_snapshots,
                             share_server=None):
        """Update the replica_state of a replica."""
        return self.plugin.update_replica_state(context,
                                                replica_list,
                                                replica,
                                                access_rules,
                                                replica_snapshots,
                                                share_server)

    def promote_replica(self, context, replica_list, replica, access_rules,
                        share_server=None):
        """Promote a replica to 'active' replica state.."""
        return self.plugin.promote_replica(context,
                                           replica_list,
                                           replica,
                                           access_rules,
                                           share_server)

    def delete_replica(self, context, replica_list, replica_snapshots,
                       replica, share_server=None):
        """Delete a replica."""
        self.plugin.delete_replica(context,
                                   replica_list,
                                   replica_snapshots,
                                   replica,
                                   share_server)

    def revert_to_snapshot(self, context, snapshot, share_access_rules,
                           snapshot_access_rules, share_server=None):
        self.plugin.revert_to_snapshot(context,
                                       snapshot,
                                       share_access_rules,
                                       snapshot_access_rules,
                                       share_server)



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests For AvailabilityZoneFilter.
"""
import ddt
from oslo_context import context

from manila.scheduler.filters import availability_zone
from manila import test
from manila.tests.scheduler import fakes


@ddt.ddt
class HostFiltersTestCase(test.TestCase):
    """Test case for AvailabilityZoneFilter."""

    def setUp(self):
        super(HostFiltersTestCase, self).setUp()
        self.filter = availability_zone.AvailabilityZoneFilter()
        self.az_id = 'e3ecad6f-e984-4cd1-b149-d83c962374a8'
        self.fake_service = {
            'service': {
                'availability_zone_id': self.az_id,
                'availability_zone': {
                    'name': 'nova',
                    'id': self.az_id
                }
            }
        }

    @staticmethod
    def _make_zone_request(zone, is_admin=False):
        ctxt = context.RequestContext('fake', 'fake', is_admin=is_admin)
        return {
            'context': ctxt,
            'request_spec': {
                'resource_properties': {
                    'availability_zone_id': zone
                }
            }
        }

    def test_availability_zone_filter_same(self):
        request = self._make_zone_request(self.az_id)
        host = fakes.FakeHostState('host1', self.fake_service)
        self.assertTrue(self.filter.host_passes(host, request))

    def test_availability_zone_filter_different(self):
        request = self._make_zone_request('bad')
        host = fakes.FakeHostState('host1', self.fake_service)
        self.assertFalse(self.filter.host_passes(host, request))

    def test_availability_zone_filter_empty(self):
        request = {}
        host = fakes.FakeHostState('host1', self.fake_service)
        self.assertTrue(self.filter.host_passes(host, request))

    def test_availability_zone_filter_both_request_AZ_and_type_AZs_match(self):
        request = self._make_zone_request(
            '9382098d-d40f-42a2-8f31-8eb78ee18c02')
        request['request_spec']['availability_zones'] = [
            'nova', 'super nova', 'hypernova']
        service = {
            'availability_zone': {
                'name': 'nova',
                'id': '9382098d-d40f-42a2-8f31-8eb78ee18c02',
            },
            'availability_zone_id': '9382098d-d40f-42a2-8f31-8eb78ee18c02',
        }
        host = fakes.FakeHostState('host1', {'service': service})

        self.assertTrue(self.filter.host_passes(host, request))

    @ddt.data((['zone1', 'zone2', 'zone 4', 'zone3'], 'zone2', True),
              (['zone1zone2zone3'], 'zone2', False),
              (['zone1zone2zone3'], 'nova', False),
              (['zone1', 'zone2', 'zone 4', 'zone3'], 'zone 4', True))
    @ddt.unpack
    def test_availability_zone_filter_only_share_type_AZs(
            self, supported_azs, request_az, host_passes):
        service = {
            'availability_zone': {
                'name': request_az,
                'id': '9382098d-d40f-42a2-8f31-8eb78ee18c02',
            },
            'availability_zone_id': '9382098d-d40f-42a2-8f31-8eb78ee18c02',
        }
        request = self._make_zone_request(None)
        request['request_spec']['availability_zones'] = supported_azs
        host = fakes.FakeHostState('host1', {'service': service})

        self.assertEqual(host_passes, self.filter.host_passes(host, request))



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# Copyright (c) 2016 EMC Corporation
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log
from oslo_utils import strutils

from manila.scheduler.filters import extra_specs_ops

LOG = log.getLogger(__name__)


def generate_stats(host_state, properties):
    """Generates statistics from host and share data."""

    host_stats = {
        'host': host_state.host,
        'share_backend_name': host_state.share_backend_name,
        'vendor_name': host_state.vendor_name,
        'driver_version': host_state.driver_version,
        'storage_protocol': host_state.storage_protocol,
        'qos': host_state.qos,
        'total_capacity_gb': host_state.total_capacity_gb,
        'allocated_capacity_gb': host_state.allocated_capacity_gb,
        'free_capacity_gb': host_state.free_capacity_gb,
        'reserved_percentage': host_state.reserved_percentage,
        'driver_handles_share_servers':
            host_state.driver_handles_share_servers,
        'thin_provisioning': host_state.thin_provisioning,
        'updated': host_state.updated,
        'dedupe': host_state.dedupe,
        'compression': host_state.compression,
        'snapshot_support': host_state.snapshot_support,
        'create_share_from_snapshot_support':
            host_state.create_share_from_snapshot_support,
        'revert_to_snapshot_support': host_state.revert_to_snapshot_support,
        'mount_snapshot_support': host_state.mount_snapshot_support,
        'replication_domain': host_state.replication_domain,
        'replication_type': host_state.replication_type,
        'provisioned_capacity_gb': host_state.provisioned_capacity_gb,
        'pools': host_state.pools,
        'max_over_subscription_ratio':
            host_state.max_over_subscription_ratio,
        'sg_consistent_snapshot_support': (
            host_state.sg_consistent_snapshot_support),
        'ipv4_support': host_state.ipv4_support,
        'ipv6_support': host_state.ipv6_support,
    }

    host_caps = host_state.capabilities

    share_type = properties.get('share_type', {})
    extra_specs = share_type.get('extra_specs', {})

    share_group_type = properties.get('share_group_type', {})
    group_specs = share_group_type.get('group_specs', {})

    request_spec = properties.get('request_spec', {})
    share_stats = request_spec.get('resource_properties', {})

    stats = {
        'host_stats': host_stats,
        'host_caps': host_caps,
        'share_type': share_type,
        'extra_specs': extra_specs,
        'share_stats': share_stats,
        'share_group_type': share_group_type,
        'group_specs': group_specs,
    }

    return stats


def use_thin_logic(share_type):
    # NOTE(xyang): To preserve the existing behavior, we use thin logic
    # to evaluate in two cases:
    # 1) 'thin_provisioning' is not set in extra specs (This is for
    #    backward compatibility. If not set, the scheduler behaves
    #    the same as before this bug fix).
    # 2) 'thin_provisioning' is set in extra specs and it is
    #    '<is> True' or 'True'.
    # Otherwise we use the thick logic to evaluate.
    use_thin_logic = True
    thin_spec = None
    try:
        thin_spec = share_type.get('extra_specs', {}).get(
            'thin_provisioning')
        if thin_spec is None:
            thin_spec = share_type.get('extra_specs', {}).get(
                'capabilities:thin_provisioning')
        # NOTE(xyang) 'use_thin_logic' and 'thin_provisioning' are NOT
        # the same thing. The first purpose of "use_thin_logic" is to
        # preserve the existing scheduler behavior if 'thin_provisioning'
        # is NOT in extra_specs (if thin_spec is None, use_thin_logic
        # should be True). The second purpose of 'use_thin_logic'
        # is to honor 'thin_provisioning' if it is in extra specs (if
        # thin_spec is set to True, use_thin_logic should be True; if
        # thin_spec is set to False, use_thin_logic should be False).
        use_thin_logic = strutils.bool_from_string(
            thin_spec, strict=True) if thin_spec is not None else True
    except ValueError:
        # Check if the value of thin_spec is '<is> True'.
        if thin_spec is not None and not extra_specs_ops.match(
                True, thin_spec):
            use_thin_logic = False
    return use_thin_logic


def thin_provisioning(host_state_thin_provisioning):
    # NOTE(xyang): host_state_thin_provisioning is reported by driver.
    # It can be either bool (True or False) or
    # list ([True, False], [True], [False]).
    thin_capability = [host_state_thin_provisioning] if not isinstance(
        host_state_thin_provisioning, list) else host_state_thin_provisioning
    return True in thin_capability


def capabilities_satisfied(capabilities, extra_specs):

    # These extra-specs are not capabilities for matching hosts
    ignored_extra_specs = (
        'availability_zones', 'capabilities:availability_zones',
    )

    for key, req in extra_specs.items():
        # Ignore some extra_specs if told to
        if key in ignored_extra_specs:
            continue

        # Either not scoped format, or in capabilities scope
        scope = key.split(':')

        # Ignore scoped (such as vendor-specific) capabilities
        if len(scope) > 1 and scope[0] != "capabilities":
            continue
        # Strip off prefix if spec started with 'capabilities:'
        elif scope[0] == "capabilities":
            del scope[0]

        cap = capabilities
        for index in range(len(scope)):
            try:
                cap = cap.get(scope[index])
            except AttributeError:
                cap = None
            if cap is None:
                LOG.debug("Host doesn't provide capability '%(cap)s' "
                          "listed in the extra specs",
                          {'cap': scope[index]})
                return False

        # Make all capability values a list so we can handle lists
        cap_list = [cap] if not isinstance(cap, list) else cap

        # Loop through capability values looking for any match
        for cap_value in cap_list:
            if extra_specs_ops.match(cap_value, req):
                break
        else:
            # Nothing matched, so bail out
            LOG.debug('Share type extra spec requirement '
                      '"%(key)s=%(req)s" does not match reported '
                      'capability "%(cap)s"',
                      {'key': key, 'req': req, 'cap': cap})
            return False
    return True



# Copyright (c) 2015 Quobyte, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_config import cfg
from oslo_utils import units
import six

from manila import context
from manila import exception
from manila.share import configuration as config
from manila.share import driver
from manila.share.drivers.quobyte import jsonrpc
from manila.share.drivers.quobyte import quobyte
from manila import test
from manila.tests import fake_share

CONF = cfg.CONF


def fake_rpc_handler(name, *args, **kwargs):
    if name == 'resolveVolumeName':
        return None
    elif name == 'createVolume':
        return {'volume_uuid': 'voluuid'}
    elif name == 'exportVolume':
        return {'nfs_server_ip': 'fake_location',
                'nfs_export_path': '/fake_share'}
    elif name == 'getConfiguration':
        return {
            "tenant_configuration": [{
                "domain_name": "fake_domain_name",
                "volume_access": [
                    {"volume_uuid": "fake_id_1",
                     "restrict_to_network": "10.0.0.1",
                     "read_only": False},
                    {"volume_uuid": "fake_id_1",
                     "restrict_to_network": "10.0.0.2",
                     "read_only": False},
                    {"volume_uuid": "fake_id_2",
                     "restrict_to_network": "10.0.0.3",
                     "read_only": False}
                ]},
                {"domain_name": "fake_domain_name_2",
                 "volume_access": [
                     {"volume_uuid": "fake_id_3",
                      "restrict_to_network": "10.0.0.4",
                      "read_only": False},
                     {"volume_uuid": "fake_id_3",
                      "restrict_to_network": "10.0.0.5",
                      "read_only": True},
                     {"volume_uuid": "fake_id_4",
                      "restrict_to_network": "10.0.0.6",
                      "read_only": False}
                 ]}
            ]
        }
    else:
        return "Unknown fake rpc handler call"


def create_fake_access(access_adr,
                       access_id='fake_access_id',
                       access_type='ip',
                       access_level='rw'):
    return {
        'access_id': access_id,
        'access_type': access_type,
        'access_to': access_adr,
        'access_level': access_level
    }


class QuobyteShareDriverTestCase(test.TestCase):
    """Tests QuobyteShareDriver."""

    def setUp(self):
        super(QuobyteShareDriverTestCase, self).setUp()

        self._context = context.get_admin_context()

        CONF.set_default('driver_handles_share_servers', False)

        self.fake_conf = config.Configuration(None)
        self._driver = quobyte.QuobyteShareDriver(configuration=self.fake_conf)
        self._driver.rpc = mock.Mock()
        self.share = fake_share.fake_share(
            share_proto='NFS',
            export_location='fake_location:/quobyte/fake_share')
        self.access = fake_share.fake_access()

    @mock.patch('manila.share.drivers.quobyte.jsonrpc.JsonRpc', mock.Mock())
    def test_do_setup_success(self):
        self._driver.rpc.call = mock.Mock(return_value=None)

        self._driver.do_setup(self._context)

        self._driver.rpc.call.assert_called_with('getInformation', {})

    @mock.patch('manila.share.drivers.quobyte.jsonrpc.JsonRpc.__init__',
                mock.Mock(return_value=None))
    @mock.patch.object(jsonrpc.JsonRpc, 'call',
                       side_effect=exception.QBRpcException)
    def test_do_setup_failure(self, mock_call):
        self.assertRaises(exception.QBException,
                          self._driver.do_setup, self._context)

    @mock.patch.object(quobyte.QuobyteShareDriver, "_resize_share")
    def test_create_share_new_volume(self, qb_resize_mock):
        self._driver.rpc.call = mock.Mock(wraps=fake_rpc_handler)

        result = self._driver.create_share(self._context, self.share)

        self.assertEqual(self.share['export_location'], result)
        self._driver.rpc.call.assert_has_calls([
            mock.call('createVolume', dict(
                name=self.share['name'],
                tenant_domain=self.share['project_id'],
                root_user_id=self.fake_conf.quobyte_default_volume_user,
                root_group_id=self.fake_conf.quobyte_default_volume_group,
                configuration_name=self.fake_conf.quobyte_volume_configuration
            )),
            mock.call('exportVolume',
                      dict(protocol='NFS', volume_uuid='voluuid'))])
        qb_resize_mock.assert_called_once_with(self.share, self.share['size'])

    @mock.patch.object(quobyte.QuobyteShareDriver, "_resize_share")
    def test_create_share_existing_volume(self, qb_resize_mock):
        self._driver.rpc.call = mock.Mock(wraps=fake_rpc_handler)

        result = self._driver.create_share(self._context, self.share)

        self.assertEqual(self.share['export_location'], result)
        resolv_params = {'tenant_domain': 'fake_project_uuid',
                         'volume_name': 'fakename'}
        sett_params = {'tenant': {'tenant_id': 'fake_project_uuid'}}
        create_params = dict(
            name='fakename',
            tenant_domain='fake_project_uuid',
            root_user_id='root',
            root_group_id='root',
            configuration_name='BASE')
        self._driver.rpc.call.assert_has_calls([
            mock.call('resolveVolumeName', resolv_params,
                      [jsonrpc.ERROR_ENOENT, jsonrpc.ERROR_ENTITY_NOT_FOUND]),
            mock.call('setTenant', sett_params,
                      expected_errors=[jsonrpc.ERROR_GARBAGE_ARGS]),
            mock.call('createVolume', create_params),
            mock.call('exportVolume', dict(protocol='NFS',
                                           volume_uuid='voluuid'))])
        qb_resize_mock.assert_called_once_with(self.share, self.share['size'])

    def test_create_share_wrong_protocol(self):
        share = {'share_proto': 'WRONG_PROTOCOL'}

        self.assertRaises(exception.QBException,
                          self._driver.create_share,
                          context=None,
                          share=share)

    def test_delete_share_existing_volume(self):
        def rpc_handler(name, *args):
            if name == 'resolveVolumeName':
                return {'volume_uuid': 'voluuid'}
            elif name == 'exportVolume':
                return {}

        self._driver.configuration.quobyte_delete_shares = True
        self._driver.rpc.call = mock.Mock(wraps=rpc_handler)

        self._driver.delete_share(self._context, self.share)

        resolv_params = {'volume_name': 'fakename',
                         'tenant_domain': 'fake_project_uuid'}
        self._driver.rpc.call.assert_has_calls([
            mock.call('resolveVolumeName', resolv_params,
                      [jsonrpc.ERROR_ENOENT, jsonrpc.ERROR_ENTITY_NOT_FOUND]),
            mock.call('deleteVolume', {'volume_uuid': 'voluuid'})])

    def test_delete_share_existing_volume_disabled(self):
        def rpc_handler(name, *args):
            if name == 'resolveVolumeName':
                return {'volume_uuid': 'voluuid'}
            elif name == 'exportVolume':
                return {}

        CONF.set_default('quobyte_delete_shares', False)
        self._driver.rpc.call = mock.Mock(wraps=rpc_handler)

        self._driver.delete_share(self._context, self.share)

        self._driver.rpc.call.assert_called_with(
            'exportVolume', {'volume_uuid': 'voluuid', 'remove_export': True})

    @mock.patch.object(quobyte.LOG, 'warning')
    def test_delete_share_nonexisting_volume(self, mock_warning):
        def rpc_handler(name, *args):
            if name == 'resolveVolumeName':
                return None

        self._driver.rpc.call = mock.Mock(wraps=rpc_handler)

        self._driver.delete_share(self._context, self.share)

        mock_warning.assert_called_with(
            'No volume found for share %(project_id)s/%(name)s',
            {'project_id': 'fake_project_uuid', 'name': 'fakename'})

    def test_allow_access(self):
        def rpc_handler(name, *args):
            if name == 'resolveVolumeName':
                return {'volume_uuid': 'voluuid'}
            elif name == 'exportVolume':
                return {'nfs_server_ip': '10.10.1.1',
                        'nfs_export_path': '/voluuid'}

        self._driver.rpc.call = mock.Mock(wraps=rpc_handler)

        self._driver._allow_access(self._context, self.share, self.access)

        exp_params = {'volume_uuid': 'voluuid',
                      'read_only': False,
                      'add_allow_ip': '10.0.0.1'}
        self._driver.rpc.call.assert_called_with('exportVolume', exp_params)

    def test_allow_ro_access(self):
        def rpc_handler(name, *args):
            if name == 'resolveVolumeName':
                return {'volume_uuid': 'voluuid'}
            elif name == 'exportVolume':
                return {'nfs_server_ip': '10.10.1.1',
                        'nfs_export_path': '/voluuid'}

        self._driver.rpc.call = mock.Mock(wraps=rpc_handler)
        ro_access = fake_share.fake_access(access_level='ro')

        self._driver._allow_access(self._context, self.share, ro_access)

        exp_params = {'volume_uuid': 'voluuid',
                      'read_only': True,
                      'add_allow_ip': '10.0.0.1'}
        self._driver.rpc.call.assert_called_with('exportVolume', exp_params)

    def test_allow_access_nonip(self):
        self._driver.rpc.call = mock.Mock(wraps=fake_rpc_handler)

        self.access = fake_share.fake_access(**{"access_type":
                                                "non_existant_access_type"})

        self.assertRaises(exception.InvalidShareAccess,
                          self._driver._allow_access,
                          self._context, self.share, self.access)

    def test_deny_access(self):
        def rpc_handler(name, *args):
            if name == 'resolveVolumeName':
                return {'volume_uuid': 'voluuid'}
            elif name == 'exportVolume':
                return {'nfs_server_ip': '10.10.1.1',
                        'nfs_export_path': '/voluuid'}

        self._driver.rpc.call = mock.Mock(wraps=rpc_handler)

        self._driver._deny_access(self._context, self.share, self.access)

        self._driver.rpc.call.assert_called_with(
            'exportVolume',
            {'volume_uuid': 'voluuid', 'remove_allow_ip': '10.0.0.1'})

    @mock.patch.object(quobyte.LOG, 'debug')
    def test_deny_access_nonip(self, mock_debug):
        self._driver.rpc.call = mock.Mock(wraps=fake_rpc_handler)
        self.access = fake_share.fake_access(
            access_type="non_existant_access_type")

        self._driver._deny_access(self._context, self.share, self.access)

        mock_debug.assert_called_with(
            'Quobyte driver only supports ip access control. '
            'Ignoring deny access call for %s , %s',
            'fakename', 'fake_project_uuid')

    def test_resolve_volume_name(self):
        self._driver.rpc.call = mock.Mock(
            return_value={'volume_uuid': 'fake_uuid'})

        self._driver._resolve_volume_name('fake_vol_name', 'fake_domain_name')

        exp_params = {'volume_name': 'fake_vol_name',
                      'tenant_domain': 'fake_domain_name'}
        self._driver.rpc.call.assert_called_with(
            'resolveVolumeName', exp_params,
            [jsonrpc.ERROR_ENOENT, jsonrpc.ERROR_ENTITY_NOT_FOUND])

    def test_resolve_volume_name_NOENT(self):
        self._driver.rpc.call = mock.Mock(
            return_value=None)

        self.assertIsNone(
            self._driver._resolve_volume_name('fake_vol_name',
                                              'fake_domain_name'))
        self._driver.rpc.call.assert_called_once_with(
            'resolveVolumeName',
            dict(volume_name='fake_vol_name',
                 tenant_domain='fake_domain_name'),
            [jsonrpc.ERROR_ENOENT, jsonrpc.ERROR_ENTITY_NOT_FOUND]
        )

    def test_resolve_volume_name_other_error(self):
        self._driver.rpc.call = mock.Mock(
            side_effect=exception.QBRpcException(
                result='fubar',
                qbcode=666))

        self.assertRaises(exception.QBRpcException,
                          self._driver._resolve_volume_name,
                          volume_name='fake_vol_name',
                          tenant_domain='fake_domain_name')

    @mock.patch.object(driver.ShareDriver, '_update_share_stats')
    def test_update_share_stats(self, mock_uss):
        self._driver._get_capacities = mock.Mock(return_value=[42, 23])

        self._driver._update_share_stats()

        mock_uss.assert_called_once_with(
            dict(storage_protocol='NFS',
                 vendor_name='Quobyte',
                 share_backend_name=self._driver.backend_name,
                 driver_version=self._driver.DRIVER_VERSION,
                 total_capacity_gb=42,
                 free_capacity_gb=23,
                 reserved_percentage=0))

    def test_get_capacities_gb(self):
        capval = 42115548133
        useval = 19695128917
        replfact = 3
        self._driver._get_qb_replication_factor = mock.Mock(
            return_value=replfact)
        self._driver.rpc.call = mock.Mock(
            return_value={'total_physical_capacity': six.text_type(capval),
                          'total_physical_usage': six.text_type(useval)})

        self.assertEqual((39.223160718, 6.960214182),
                         self._driver._get_capacities())

    def test_get_capacities_gb_full(self):
        capval = 1024 * 1024 * 1024 * 3
        useval = 1024 * 1024 * 1024 * 3 + 1
        replfact = 1
        self._driver._get_qb_replication_factor = mock.Mock(
            return_value=replfact)
        self._driver.rpc.call = mock.Mock(
            return_value={'total_physical_capacity': six.text_type(capval),
                          'total_physical_usage': six.text_type(useval)})

        self.assertEqual((3.0, 0), self._driver._get_capacities())

    def test_get_replication(self):
        fakerepl = 42
        self._driver.configuration.quobyte_volume_configuration = 'fakeVolConf'
        self._driver.rpc.call = mock.Mock(
            return_value={'configuration':
                          {'volume_metadata_configuration':
                           {'replication_factor':
                            six.text_type(fakerepl)}}})

        self.assertEqual(fakerepl, self._driver._get_qb_replication_factor())

    @mock.patch.object(quobyte.QuobyteShareDriver,
                       "_resolve_volume_name",
                       return_value="fake_uuid")
    def test_ensure_share(self, mock_qb_resolve_volname):
        self._driver.rpc.call = mock.Mock(wraps=fake_rpc_handler)

        result = self._driver.ensure_share(self._context, self.share, None)

        self.assertEqual(self.share["export_location"], result)
        (mock_qb_resolve_volname.
         assert_called_once_with(self.share['name'],
                                 self.share['project_id']))
        self._driver.rpc.call.assert_has_calls([
            mock.call('exportVolume', dict(
                volume_uuid="fake_uuid",
                protocol='NFS'
            ))])

    @mock.patch.object(quobyte.QuobyteShareDriver,
                       "_resolve_volume_name",
                       return_value=None)
    def test_ensure_deleted_share(self, mock_qb_resolve_volname):
        self._driver.rpc.call = mock.Mock(wraps=fake_rpc_handler)

        self.assertRaises(exception.ShareResourceNotFound,
                          self._driver.ensure_share,
                          self._context, self.share, None)
        (mock_qb_resolve_volname.
         assert_called_once_with(self.share['name'],
                                 self.share['project_id']))

    @mock.patch.object(quobyte.QuobyteShareDriver, "_resize_share")
    def test_extend_share(self, mock_qsd_resize_share):
        self._driver.extend_share(ext_share=self.share,
                                  ext_size=2,
                                  share_server=None)
        mock_qsd_resize_share.assert_called_once_with(share=self.share,
                                                      new_size=2)

    @mock.patch.object(quobyte.QuobyteShareDriver, "_resolve_volume_name",
                       return_value="fake_volume_uuid")
    def test_resize_share(self, mock_qb_resolv):
        self._driver.rpc.call = mock.Mock(wraps=fake_rpc_handler)
        manila_size = 7
        newsize_bytes = manila_size * units.Gi

        self._driver._resize_share(share=self.share, new_size=manila_size)

        exp_params = {
            "quotas": [{
                "consumer": [{
                    "type": "VOLUME",
                    "identifier": "fake_volume_uuid",
                    "tenant_id": self.share["project_id"]
                }],
                "limits": [{
                    "type": "LOGICAL_DISK_SPACE",
                    "value": newsize_bytes,
                }],
            }]}
        self._driver.rpc.call.assert_has_calls([
            mock.call('setQuota', exp_params)])
        mock_qb_resolv.assert_called_once_with(self.share['name'],
                                               self.share['project_id'])

    @mock.patch.object(quobyte.QuobyteShareDriver,
                       "_resolve_volume_name",
                       return_value="fake_id_3")
    def test_fetch_existing_access(self, mock_qb_resolve_volname):
        self._driver.rpc.call = mock.Mock(wraps=fake_rpc_handler)
        old_access_1 = create_fake_access(access_id="old_1",
                                          access_adr="10.0.0.4")
        old_access_2 = create_fake_access(access_id="old_2",
                                          access_adr="10.0.0.5")

        exist_list = self._driver._fetch_existing_access(context=self._context,
                                                         share=self.share)

        # assert expected result here
        self.assertEqual([old_access_1['access_to'],
                          old_access_2['access_to']],
                         [e.get('access_to') for e in exist_list])
        (mock_qb_resolve_volname.
         assert_called_once_with(self.share['name'],
                                 self.share['project_id']))

    @mock.patch.object(quobyte.QuobyteShareDriver, "_resize_share")
    def test_shrink_share(self, mock_qsd_resize_share):
        self._driver.shrink_share(shrink_share=self.share,
                                  shrink_size=3,
                                  share_server=None)
        mock_qsd_resize_share.assert_called_once_with(share=self.share,
                                                      new_size=3)

    def test_subtract_access_lists(self):
        access_1 = create_fake_access(access_id="new_1",
                                      access_adr="10.0.0.5",
                                      access_type="rw",)
        access_2 = create_fake_access(access_id="old_1",
                                      access_adr="10.0.0.1",
                                      access_type="rw")
        access_3 = create_fake_access(access_id="old_2",
                                      access_adr="10.0.0.3",
                                      access_type="ro")
        access_4 = create_fake_access(access_id="new_2",
                                      access_adr="10.0.0.6",
                                      access_type="rw")
        access_5 = create_fake_access(access_id="old_3",
                                      access_adr="10.0.0.4",
                                      access_type="rw")
        min_list = [access_1, access_2, access_3, access_4]
        sub_list = [access_5, access_3, access_2]

        self.assertEqual([access_1, access_4],
                         self._driver._subtract_access_lists(min_list,
                                                             sub_list))

    def test_subtract_access_lists_level(self):
        access_1 = create_fake_access(access_id="new_1",
                                      access_adr="10.0.0.5",
                                      access_level="rw")
        access_2 = create_fake_access(access_id="old_1",
                                      access_adr="10.0.0.1",
                                      access_level="rw")
        access_3 = create_fake_access(access_id="old_2",
                                      access_adr="10.0.0.3",
                                      access_level="rw")
        access_4 = create_fake_access(access_id="new_2",
                                      access_adr="10.0.0.6",
                                      access_level="rw")
        access_5 = create_fake_access(access_id="old_2_ro",
                                      access_adr="10.0.0.3",
                                      access_level="ro")
        min_list = [access_1, access_2, access_3, access_4]
        sub_list = [access_5, access_2]

        self.assertEqual([access_1, access_3, access_4],
                         self._driver._subtract_access_lists(min_list,
                                                             sub_list))

    def test_subtract_access_lists_type(self):
        access_1 = create_fake_access(access_id="new_1",
                                      access_adr="10.0.0.5",
                                      access_type="ip")
        access_2 = create_fake_access(access_id="old_1",
                                      access_adr="10.0.0.1",
                                      access_type="ip")
        access_3 = create_fake_access(access_id="old_2",
                                      access_adr="10.0.0.3",
                                      access_type="ip")
        access_4 = create_fake_access(access_id="new_2",
                                      access_adr="10.0.0.6",
                                      access_type="ip")
        access_5 = create_fake_access(access_id="old_2_ro",
                                      access_adr="10.0.0.3",
                                      access_type="other")
        min_list = [access_1, access_2, access_3, access_4]
        sub_list = [access_5, access_2]

        self.assertEqual([access_1, access_3, access_4],
                         self._driver._subtract_access_lists(min_list,
                                                             sub_list))

    @mock.patch.object(quobyte.QuobyteShareDriver, "_allow_access")
    @mock.patch.object(quobyte.QuobyteShareDriver, "_deny_access")
    def test_update_access_add_delete(self, qb_deny_mock, qb_allow_mock):
        access_1 = create_fake_access(access_id="new_1",
                                      access_adr="10.0.0.5",
                                      access_level="rw")
        access_2 = create_fake_access(access_id="old_1",
                                      access_adr="10.0.0.1",
                                      access_level="rw")
        access_3 = create_fake_access(access_id="old_2",
                                      access_adr="10.0.0.3",
                                      access_level="rw")

        self._driver.update_access(self._context,
                                   self.share,
                                   access_rules=None,
                                   add_rules=[access_1],
                                   delete_rules=[access_2, access_3])

        qb_allow_mock.assert_called_once_with(self._context,
                                              self.share, access_1)
        deny_calls = [mock.call(self._context, self.share, access_2),
                      mock.call(self._context, self.share, access_3)]
        qb_deny_mock.assert_has_calls(deny_calls)

    @mock.patch.object(quobyte.LOG, "warning")
    def test_update_access_no_rules(self, qb_log_mock):
        self._driver.update_access(context=None, share=None, access_rules=[],
                                   add_rules=[], delete_rules=[])

        qb_log_mock.assert_has_calls([mock.ANY])

    @mock.patch.object(quobyte.QuobyteShareDriver, "_subtract_access_lists")
    @mock.patch.object(quobyte.QuobyteShareDriver, "_fetch_existing_access")
    @mock.patch.object(quobyte.QuobyteShareDriver, "_allow_access")
    def test_update_access_recovery_additionals(self,
                                                qb_allow_mock,
                                                qb_exist_mock,
                                                qb_subtr_mock):
        new_access_1 = create_fake_access(access_id="new_1",
                                          access_adr="10.0.0.2")
        old_access = create_fake_access(access_id="fake_access_id",
                                        access_adr="10.0.0.1")
        new_access_2 = create_fake_access(access_id="new_2",
                                          access_adr="10.0.0.3")
        add_access_rules = [new_access_1,
                            old_access,
                            new_access_2]
        qb_exist_mock.return_value = [old_access]
        qb_subtr_mock.side_effect = [[new_access_1, new_access_2], []]

        self._driver.update_access(self._context, self.share,
                                   access_rules=add_access_rules, add_rules=[],
                                   delete_rules=[])

        assert_calls = [mock.call(self._context, self.share, new_access_1),
                        mock.call(self._context, self.share, new_access_2)]
        qb_allow_mock.assert_has_calls(assert_calls, any_order=True)
        qb_exist_mock.assert_called_once_with(self._context, self.share)

    @mock.patch.object(quobyte.QuobyteShareDriver, "_subtract_access_lists")
    @mock.patch.object(quobyte.QuobyteShareDriver, "_fetch_existing_access")
    @mock.patch.object(quobyte.QuobyteShareDriver, "_deny_access")
    def test_update_access_recovery_superfluous(self,
                                                qb_deny_mock,
                                                qb_exist_mock,
                                                qb_subtr_mock):

        old_access_1 = create_fake_access(access_id="old_1",
                                          access_adr="10.0.0.1")
        missing_access_1 = create_fake_access(access_id="mis_1",
                                              access_adr="10.0.0.2")
        old_access_2 = create_fake_access(access_id="old_2",
                                          access_adr="10.0.0.3")
        qb_exist_mock.side_effect = [[old_access_1, old_access_2]]
        qb_subtr_mock.side_effect = [[], [missing_access_1]]
        old_access_rules = [old_access_1, old_access_2]

        self._driver.update_access(self._context, self.share,
                                   access_rules=old_access_rules, add_rules=[],
                                   delete_rules=[])

        qb_deny_mock.assert_called_once_with(self._context,
                                             self.share,
                                             (missing_access_1))
        qb_exist_mock.assert_called_once_with(self._context, self.share)

    @mock.patch.object(quobyte.QuobyteShareDriver, "_subtract_access_lists")
    @mock.patch.object(quobyte.QuobyteShareDriver, "_fetch_existing_access")
    @mock.patch.object(quobyte.QuobyteShareDriver, "_deny_access")
    @mock.patch.object(quobyte.QuobyteShareDriver, "_allow_access")
    def test_update_access_recovery_add_superfluous(self,
                                                    qb_allow_mock,
                                                    qb_deny_mock,
                                                    qb_exist_mock,
                                                    qb_subtr_mock):
        new_access_1 = create_fake_access(access_id="new_1",
                                          access_adr="10.0.0.5")
        old_access_1 = create_fake_access(access_id="old_1",
                                          access_adr="10.0.0.1")
        old_access_2 = create_fake_access(access_id="old_2",
                                          access_adr="10.0.0.3")
        old_access_3 = create_fake_access(access_id="old_3",
                                          access_adr="10.0.0.4")
        miss_access_1 = create_fake_access(access_id="old_3",
                                           access_adr="10.0.0.4")
        new_access_2 = create_fake_access(access_id="new_2",
                                          access_adr="10.0.0.3",
                                          access_level="ro")
        new_access_rules = [new_access_1, old_access_1, old_access_2,
                            old_access_3, new_access_2]
        qb_exist_mock.return_value = [old_access_1, old_access_2,
                                      old_access_3, miss_access_1]
        qb_subtr_mock.side_effect = [[new_access_1, new_access_2],
                                     [miss_access_1, old_access_2]]

        self._driver.update_access(self._context, self.share,
                                   new_access_rules, add_rules=[],
                                   delete_rules=[])

        a_calls = [mock.call(self._context, self.share, new_access_1),
                   mock.call(self._context, self.share, new_access_2)]
        qb_allow_mock.assert_has_calls(a_calls)
        b_calls = [mock.call(self._context, self.share, miss_access_1),
                   mock.call(self._context, self.share, old_access_2)]
        qb_deny_mock.assert_has_calls(b_calls)
        qb_exist_mock.assert_called_once_with(self._context, self.share)



# Copyright (c) 2017 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

RULE_ADMIN_OR_OWNER = 'rule:admin_or_owner'
RULE_ADMIN_API = 'rule:admin_api'
RULE_DEFAULT = 'rule:default'

rules = [
    policy.RuleDefault(name='context_is_admin', check_str='role:admin'),
    policy.RuleDefault(
        name='admin_or_owner',
        check_str='is_admin:True or project_id:%(project_id)s'),
    policy.RuleDefault(name='default', check_str=RULE_ADMIN_OR_OWNER),
    policy.RuleDefault(name='admin_api', check_str='is_admin:True'),
]


def list_rules():
    return rules



# Copyright 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import with_statement

from alembic import context

from manila.db.sqlalchemy import api as db_api
from manila.db.sqlalchemy import models as db_models


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.
    """
    engine = db_api.get_engine()
    connection = engine.connect()
    target_metadata = db_models.ManilaBase.metadata
    # pylint: disable=no-member
    context.configure(connection=connection,
                      target_metadata=target_metadata)
    try:
        with context.begin_transaction():
            context.run_migrations()
    finally:
        connection.close()


run_migrations_online()



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_policy import opts
from oslo_service import wsgi

from manila.common import config

CONF = config.CONF


def set_defaults(conf):
    _safe_set_of_opts(conf, 'verbose', True)
    _safe_set_of_opts(conf, 'state_path', os.path.abspath(
        os.path.join(os.path.dirname(__file__),
                     '..',
                     '..')))
    _safe_set_of_opts(conf, 'connection', "sqlite://", group='database')
    _safe_set_of_opts(conf, 'sqlite_synchronous', False)
    _POLICY_PATH = os.path.abspath(os.path.join(CONF.state_path,
                                                'manila/tests/policy.json'))
    opts.set_defaults(conf, policy_file=_POLICY_PATH)
    _safe_set_of_opts(conf, 'share_export_ip', '0.0.0.0')
    _safe_set_of_opts(conf, 'service_instance_user', 'fake_user')
    _API_PASTE_PATH = os.path.abspath(os.path.join(CONF.state_path,
                                                   'etc/manila/api-paste.ini'))
    wsgi.register_opts(conf)
    _safe_set_of_opts(conf, 'api_paste_config', _API_PASTE_PATH)
    _safe_set_of_opts(conf, 'share_driver',
                      'manila.tests.fake_driver.FakeShareDriver')
    _safe_set_of_opts(conf, 'auth_strategy', 'noauth')

    _safe_set_of_opts(conf, 'zfs_share_export_ip', '1.1.1.1')
    _safe_set_of_opts(conf, 'zfs_service_ip', '2.2.2.2')
    _safe_set_of_opts(conf, 'zfs_zpool_list', ['foo', 'bar'])
    _safe_set_of_opts(conf, 'zfs_share_helpers', 'NFS=foo.bar.Helper')
    _safe_set_of_opts(conf, 'zfs_replica_snapshot_prefix', 'foo_prefix_')

    _safe_set_of_opts(conf, 'hitachi_hsp_host', '172.24.47.190')
    _safe_set_of_opts(conf, 'hitachi_hsp_username', 'hsp_user')
    _safe_set_of_opts(conf, 'hitachi_hsp_password', 'hsp_password')

    _safe_set_of_opts(conf, 'as13000_nas_ip', '1.1.1.1')
    _safe_set_of_opts(conf, 'as13000_nas_login', 'admin')
    _safe_set_of_opts(conf, 'as13000_nas_password', 'password')
    _safe_set_of_opts(conf, 'as13000_share_pools', 'pool0')

    _safe_set_of_opts(conf, 'qnap_management_url', 'http://1.2.3.4:8080')
    _safe_set_of_opts(conf, 'qnap_share_ip', '1.2.3.4')
    _safe_set_of_opts(conf, 'qnap_nas_login', 'admin')
    _safe_set_of_opts(conf, 'qnap_nas_password', 'qnapadmin')
    _safe_set_of_opts(conf, 'qnap_poolname', 'Storage Pool 1')

    _safe_set_of_opts(conf, 'unity_server_meta_pool', 'nas_server_pool')


def _safe_set_of_opts(conf, *args, **kwargs):
    try:
        conf.set_default(*args, **kwargs)
    except config.cfg.NoSuchOptError:
        # Assumed that opt is not imported and not used
        pass
